import wntr
import toolkit_constants
import numpy as np
import matplotlib.pyplot as plt
import os


class Epanet_Except(Exception):
    pass


class ID_Exception(Exception):
    pass


class epanet:

    def __init__(self, inp_file):
        self._d = wntr.epanet.toolkit.ENepanet()

        for root, dirs, files in os.walk(os.path.dirname(os.path.abspath(inp_file))):
            for name in files:
                if name == inp_file:
                    inp_file = os.path.join(root, inp_file)
                    break
            else:
                continue
            break

        self._d.ENopen(inp_file, "temp.rpt")
        self.ToolkitConstants = toolkit_constants.toolkit_constants()
        self.morph = wntr.morph.link  
        self.elems = wntr.network.elements
        self.inp_file = inp_file
        
        # Parameters
        self.nodePressure     =  self.getNodePressure()
        self.nodeElevation    =  self.getNodeElevations()


    ############ Properties ############           
    TYPELINK = ['CVPIPE', 'PIPE', 'PUMP', 'PRV', 'PSV', 'PBV', 'FCV', 'TCV', 'GPV']
    TYPENODE = ['JUNCTION', 'RESERVOIR', 'TANK']
    TYPEPUMP = ['CONSTANT_HORSEPOWER', 'POWER_FUNCTION', 'CUSTOM']
    TYPESOURCE = ['CONCEN', 'MASS', 'SETPOINT', 'FLOWPACED']  # Constants for sources
    TYPEMIXMODEL = ['MIX1', 'MIX2', 'FIFO', 'LIFO']  # Constants for mixing models: 'MIX1', 'MIX2', 'FIFO', 'LIFO'
    TYPESTATS    = ['NONE', 'AVERAGE', 'MINIMUM', 'MAXIMUM', 'RANGE'] # Constants for statistics: 'NONE', 'AVERAGE', 'MINIMUM', 'MAXIMUM', 'RANGE'
    
    @property
    def d(self):
        return self._d
    @property
    def getControlRulesCount(self):
        """ Retrieves the number of controls.
       
        Example:
          d.getControlRulesCount
       
        See also getControls, getRuleCount.
        """
        try:
            value = self._d.ENgetcount(self.ToolkitConstants.EN_CONTROLCOUNT)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value  
    
    @property
    def getCounts(self):
        """ Retrieves the number of network components.
        Nodes, Links, Junctions, Reservoirs, Tanks, Pipes, Pumps,
        Valves, Curves, SimpleControls, RuleBasedControls, Patterns.
       
        Example:
          d.getCounts()                    # Retrieves the number of all network components
          d.getCounts().Nodes              # Retrieves the number of nodes
          d.getCounts().SimpleControls     # Retrieves the number of simple controls
       
        See also getNodeCount, getNodeJunctionCount,
                 getLinkCount, getControlRulesCount.
        """
        class value:
            def __init__(self):
                value.Nodes = []
        value = value()
        value.Nodes = self.getNodeCount()
        value.Links = self.getLinkCount()
        value.Junctions = self.getNodeJunctionCount()
        value.Reservoirs = self.getNodeReservoirCount()
        value.Tanks = self.getNodeTankCount()
        value.Pipes = self.getLinkPipeCount()
        value.Pumps = self.getLinkPumpCount()
        value.Valves = self.getLinkValveCount()
        value.Curves = self.getCurveCount()
        value.SimpleControls = self.getControlRulesCount
# value.RuleBasedControls = self.d.getRuleCount()
        value.Patterns = self.getPatternCount()     
        return value
    
    def getCurveCount(self):   
        """ Retrieves the number of curves.
       
        Example:
          d.getCurveCount()
       
        See also getCurveIndex, getCurvesInfo.
        """
        try:
            value = self.d.ENgetcount(self.ToolkitConstants.EN_CURVECOUNT)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value    
   
    def getError(self, Errcode):
        """ Retrieves the text of the message associated with a particular error or warning code.
          
        Example:
          error = 250 
          d.getError(error)
        """
        errmsg =  self.d.ENgeterror(Errcode)      
        raise Epanet_Except(errmsg)
        
    def getLinkCount(self):
        """ Retrieves the number of links.
       
        Example:
          d.getLinkCount()
       
        See also getLinkIndex, getNodeCount.
        """
        try:
            value = self.d.ENgetcount(self.ToolkitConstants.EN_LINKCOUNT)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value
    
    def getLinkQuality(self, *argv):
        """ Retrieves the value of link quality.
         Pipe quality
        
         Example 1:
            d.getLinkQuality()       # Retrieves the value of all link quality
        
         Example 2:
            d.getLinkQuality(1)      # Retrieves the value of the first link quality
        
         See also getLinkType, getLinksInfo, getLinkDiameter,
                  getLinkRoughnessCoeff, getLinkMinorLossCoeff.
         """
        return self.get_link_info(self.ToolkitConstants.EN_QUALITY, *argv)
    
    def getLinkType(self, *argv):
        """ Retrieves the link-type code for all links.
       
        Example 1:
           d.getLinkType()      # Retrieves the link-type code for all links
       
        Example 2:
           d.getLinkType(1)     # Retrieves the link-type code for the first link
       
        See also getLinkTypeIndex, getLinksInfo, getLinkDiameter,
                 getLinkLength, getLinkRoughnessCoeff, getLinkMinorLossCoeff.
        """
        try:
            lTypes = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        lTypes.append(self.TYPELINK[self.d.ENgetlinktype(i)])
                else:
                    lTypes = self.TYPELINK[self.d.ENgetlinktype(index)] 
            else:
                for i in range(self.getLinkCount()):
                    lTypes.append(self.TYPELINK[self.d.ENgetlinktype(i+1)])
        except Exception as Errcode:
            return self.getError(Errcode)
        return lTypes
    
    def getLinkTypeIndex(self, *argv):
        """ Retrieves the link-type code for all links.
       
        Example 1:
           d.getLinkTypeIndex()          # Retrieves the link-type code for all links
       
        Example 2:
           d.getLinkTypeIndex(1)         # Retrieves the link-type code for the first link
       
        Example 3:
           d.getLinkTypeIndex([2,3])     # Retrieves the link-type code for the second and third links
       
        See also getLinkType, getLinksInfo, getLinkDiameter,
                 getLinkLength, getLinkRoughnessCoeff, getLinkMinorLossCoeff.
        """
        try:
            lTypes = [] 
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        lTypes.append(self.d.ENgetlinktype(i))
                else:
                    lTypes = self.d.ENgetlinktype(index) 
            else:
                for i in range(self.getLinkCount()):
                    lTypes.append(self.d.ENgetlinktype(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return lTypes
        
    def getLinkDiameter(self, *argv):
        """ Retrieves the value of link diameters.
        Pipe/valve diameter
       
        Example 1:
           d.getLinkDiameter()      # Retrieves the value of all link diameters
       
        Example 2:
           d.getLinkDiameter(1)     # Retrieves the value of the first link diameter
       
        Example 3:
           d.getLinkDiameter([1,2])        #  Retrieves the value of the second and third link diameter
        See also getLinkType, getLinksInfo, getLinkLength,
                 getLinkRoughnessCoeff, getLinkMinorLossCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_DIAMETER, *argv)
    
    def getLinkLength(self, *argv):
        """ Retrieves the value of link lengths.
        Pipe length
       
        Example 1:
           d.getLinkLength()        # Retrieves the value of all link lengths
       
        Example 2:
           d.getLinkLength(1)       # Retrieves the value of the first link length
       
        See also getLinkType, getLinksInfo, getLinkDiameter,
                 getLinkRoughnessCoeff, getLinkMinorLossCoeff.ughnessCoeff, getLinkMinorLossCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_LENGTH, *argv)
    
    def getLinkRoughnessCoeff(self, *argv):
        """ Retrieves the value of link roughness coefficient.
        Pipe roughness coefficient
       
        Example 1:
           d.getLinkRoughnessCoeff()        # Retrieves the value of all link roughness coefficients
       
        Example 2:
           d.getLinkRoughnessCoeff(1)       # Retrieves the value of the first link roughness coefficient
       
        See also getLinkType, getLinksInfo, getLinkDiameter,
                 getLinkLength, getLinkMinorLossCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_ROUGHNESS, *argv)   

    def getLinkMinorLossCoeff(self, *argv):
        """ Retrieves the value of link minor loss coefficients.
        Pipe/valve minor loss coefficient
       
        Example 1:
          d.getLinkMinorLossCoeff()        # Retrieves the value of all link minor loss coefficients
       
        Example 2:
          d.getLinkMinorLossCoeff(1)       # Retrieves the value of the first link minor loss coefficient
       
        See also getLinkType, getLinksInfo, getLinkDiameter,
                 getLinkLength, getLinkRoughnessCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_MINORLOSS, *argv) 

    def getLinkNameID(self, *argv):
        """ Retrieves the ID label(s) of all links, or the IDs of an index set of links.
       
        Example 1:
          d.getLinkNameID()                # Retrieves the ID's of all links
       
        Example 2:
          linkIndex = 1 
          d.getLinkNameID(linkIndex)       # Retrieves the ID of the link with index = 1
       
        Example 3:
          linkIndices = [1,2,3] 
          d.getLinkNameID(linkIndices)     # Retrieves the IDs of the links with indices = 1, 2, 3
       
        See also getNodeNameID, getLinkPipeNameID, getLinkIndex.
        """
        try:
            values = [] 
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetlinkid(i))
                else:
                    values = self.d.ENgetlinkid(index) 
            else:
                for i in range(self.getLinkCount()):
                    values.append(self.d.ENgetlinkid(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return values
    
    def getLinkInitialStatus(self, *argv):
        """ Retrieves the value of all link initial status.
        Initial status (see @ref EN_LinkStatusType)
       
        Example 1:
          d.getLinkInitialStatus()        # Retrieves the value of all link initial status
       
        Example 2:
          d.getLinkInitialStatus(1)       # Retrieves the value of the first link initial status
       
        See also getLinkType, getLinksInfo, getLinkInitialSetting,
                 getLinkBulkReactionCoeff, getLinkWallReactionCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_INITSTATUS, *argv) 
    
    def getLinkInitialSetting(self, *argv):
        """ Retrieves the value of all link roughness for pipes or initial speed for pumps or initial setting for valves.
       
        Example 1:
          d.getLinkInitialSetting()       # Retrieves the value of all link initial settings
       
        Example 2:
          d.getLinkInitialSetting(1)      # Retrieves the value of the first link initial setting
       
        See also getLinkType, getLinksInfo, getLinkInitialStatus,
                 getLinkBulkReactionCoeff, getLinkWallReactionCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_INITSETTING, *argv) 
    
    def getLinkBulkReactionCoeff(self, *argv):
        """ Retrieves the value of all link bulk chemical reaction coefficient.
       
        Example 1:
          d.getLinkBulkReactionCoeff()      # Retrieves the value of all link bulk chemical reaction coefficient
       
        Example 2:
          d.getLinkBulkReactionCoeff(1)     # Retrieves the value of the first link bulk chemical reaction coefficient
       
        See also getLinkType, getLinksInfo, getLinkRoughnessCoeff,
                 getLinkMinorLossCoeff, getLinkInitialStatus,
                 getLinkInitialSetting, getLinkWallReactionCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_KBULK, *argv) 
    
    def getLinkWallReactionCoeff(self, *argv):
        """ Retrieves the value of all pipe wall chemical reaction coefficient.
       
        Example 1:
          d.getLinkWallReactionCoeff()      # Retrieves the value of all pipe wall chemical reaction coefficient
       
        Example 2:
          d.getLinkWallReactionCoeff(1)     # Retrieves the value of the first pipe wall chemical reaction coefficient
       
        See also getLinkType, getLinksInfo, getLinkRoughnessCoeff,
                 getLinkMinorLossCoeff, getLinkInitialStatus,
                 getLinkInitialSetting, getLinkBulkReactionCoeff.
        """
        return self.get_link_info(self.ToolkitConstants.EN_KWALL, *argv)  
    
    def getLinkPipeCount(self):
        """ Retrieves the number of pipes.
       
        Example:
          d.getLinkPipeCount()
       
        See also getLinkPumpCount, getLinkCount.
        """
        try:
            linkTypes = self.getLinkTypeIndex()
            value = linkTypes.count(self.ToolkitConstants.EN_CVPIPE) + linkTypes.count(self.ToolkitConstants.EN_PIPE)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value  
    
    def getLinkPumpEfficiency(self, *argv):
        """ Retrieves the current computed pump efficiency (read only). (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpEfficiency()        # Retrieves the current computed pump efficiency for all links
       
        Example 2:
          d.getLinkPumpEfficiency(1)       # Retrieves the current computed pump efficiency for the first link
       
        See also getLinkFlows, getLinkStatus, getLinkPumpState,
                 getLinkSettings, getLinkEnergy, getLinkActualQuality.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_EFFIC, *argv)
        
    def getLinkPumpCount(self):
        """ Retrieves the number of pumps.
       
        Example:
          d.getLinkPumpCount()
       
        See also getLinkPipeCount, getLinkCount.
        """
        try:
            linkTypes = self.getLinkTypeIndex()
            value = linkTypes.count(self.ToolkitConstants.EN_PUMP)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value 
    
    def getLinkPumpECost(self, *argv):
        """ Retrieves the pump average energy price. (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpECost()                # Retrieves the average energy price of all pumps
       
        Example 2:
          d.getLinkPumpECost(1)               # Retrieves the average energy price of the 1st pump
       
        Example 3:
          d = epanet('Richmond_standard.inp')
          pIndex = 950
          pIndices = d.getLinkPumpIndex()
          pindex = pIndices.index(pIndex+1)
          d.getLinkPumpECost(pindex)           # Retrieves the average energy price of the pump with link index 950
       
        See also setLinkPumpECost, getLinkPumpPower, getLinkPumpHCurve,
                 getLinkPumpEPat, getLinkPumpPatternIndex, getLinkPumpPatternNameID.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_ECOST, *argv) 

    def getLinkPumpECurve(self, *argv):
        """ Retrieves the pump efficiency v. flow curve index. (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpECurve()                # Retrieves the efficiency v. flow curve index of all pumps
       
        Example 2:
          d.getLinkPumpECurve(1)               # Retrieves the efficiency v. flow curve index of the 1st pump
       
        Example 3:
          d.getLinkPumpECurve([1,2])           # Retrieves the efficiency v. flow curve index of the first 2 pumps
       
        Example 4:
          d = epanet('Richmond_standard.inp')  # Retrieves the efficiency v. flow curve index of the pumps with lin index 950
          pIndex = 950
          pIndices = d.getLinkPumpIndex()
          pindex = pIndices.index(pIndex+1)
          d.getLinkPumpECurve(pindex)         
       
        See also setLinkPumpECurve, getLinkPumpHCurve, getLinkPumpECost,
                 getLinkPumpEPat, getLinkPumpPatternIndex, getLinkPumpPatternNameID.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_ECURVE, *argv) 

    def getLinkPumpEPat(self, *argv):
        """ Retrieves the pump energy price time pattern index. (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpEPat()                # Retrieves the energy price time pattern index of all pumps
       
        Example 2:
          d.getLinkPumpEPat(1)               # Retrieves the energy price time pattern index of the 1st pump
       
        Example 3:
          d.getLinkPumpEPat([1,2])           # Retrieves the energy price time pattern index of the first 2 pumps
       
        Example 4:
          d = epanet('Richmond_standard.inp')
          pIndex = 950
          pIndices = d.getLinkPumpIndex()
          pindex = pIndices.index(pIndex+1)
          d.getLinkPumpEPat(pindex)           # Retrieves the energy price time pattern index of pump with link index 950
       
        See also setLinkPumpEPat, getLinkPumpHCurve, getLinkPumpECurve,
                 getLinkPumpECost, getLinkPumpPatternIndex, getLinkPumpPatternNameID.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_EPAT, *argv) 


    def getLinkPumpHCurve(self, *argv):
        """ Retrieves the pump head v. flow curve index. (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpHCurve()                # Retrieves the head v. flow curve index of all pumps
       
        Example 2:
          d.getLinkPumpHCurve(1)               # Retrieves the head v. flow curve index of the 1st pump
       
        Example 3:
          d.getLinkPumpHCurve([1,2])           # Retrieves the head v. flow curve index of the first 2 pumps
       
        Example 3:
          d = epanet('Richmond_standard.inp')  # Retrieves the head v. flow curve index of pump with link index 950
          pIndex = 950
          pIndices = d.getLinkPumpIndex()
          pindex = pIndices.index(pIndex+1)
          d.getLinkPumpHCurve(pindex)         
       
        See also setLinkPumpHCurve, getLinkPumpECurve, getLinkPumpECost,
                 getLinkPumpEPat, getLinkPumpPatternIndex, getLinkPumpPatternNameID.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_HCURVE, *argv) 
        
    def getLinkPumpPatternIndex(self, *argv):
        """ Retrieves the pump speed time pattern index. (EPANET Version 2.1)
       
        Example 1: 
          d.getLinkPumpPatternIndex()                # Retrieves the speed time pattern index of all pumps
       
        Example 2:
          d.getLinkPumpPatternIndex(1)               # Retrieves the speed time pattern index of the 1st pump
       
        Example 3:
          d.getLinkPumpPatternIndex([1,2])           # Retrieves the speed time pattern index of the first 2 pumps
       
        Example 4:
          pumpIndex = d.getLinkPumpIndex 
          d.getLinkPumpPatternIndex(pumpIndex)       # Retrieves the speed time pattern index of the pumps given their indices
       
        See also setLinkPumpPatternIndex, getLinkPumpPower, getLinkPumpHCurve,
                 getLinkPumpECost, getLinkPumpEPat,  getLinkPumpPatternNameID.
        """
        value = self.get_pump_link_info(self.ToolkitConstants.EN_LINKPATTERN, *argv)
        # turn value to integer type
        if type(value) is list:
            value = [int(x) for x in value]
        else:
            value = int(value)
        return value
    
    def getLinkPumpPatternNameID(self, *argv):
        """ Retrieves pump pattern name ID. (EPANET Version 2.1)
        A value of 0 means empty
       
        Example 1:
          d.getLinkPumpPatternNameID()              # Retrieves the pattern name ID of all pumps
       
        Example 2:
          d.getLinkPumpPatternNameID(1)             # Retrieves the pattern name ID of the 1st pump
       
        Example 3:
          d.getLinkPumpPatternNameID([1,2])         # Retrieves the pattern name ID of the first 2 pumps
       
        Example 4:
          pumpIndex = d.getLinkPumpIndex 
          d.getLinkPumpPatternNameID(pumpIndex)     # Retrieves the pattern name ID of the pumps given their indices
       
        See also getLinkPumpPower, getLinkPumpHCurve, getLinkPumpECurve,
                 getLinkPumpECost, getLinkPumpEPat, getLinkPumpPatternIndex.
        """
        if len(argv) == 0:
            patindices  = self.getLinkPumpPatternIndex()
        else:
            patindices  = self.getLinkPumpPatternIndex(argv[0])
        if type(patindices) is not list:
            patindices = [patindices]
        value = []
        for i in patindices:
            if i == 0:
                value.append('')
            else:
                value.append(self.getPatternNameID(i))
        return value
    
    def getLinkPumpPower(self, *argv):
        """ Retrieves the pump constant power rating (read only). (EPANET Version 2.2)
       
        Example 1:
          d.getLinkPumpPower()                # Retrieves the constant power rating of all pumps
       
        Example 2:
          d.getLinkPumpPower(1)               # Retrieves the constant power rating of the 1st pump
       
        Example 3:
          d.getLinkPumpPower([1,2])           # Retrieves the constant power rating of the first 2 pumps
       
        Example 4:
          pumpIndex = d.getLinkPumpIndex()
          d.getLinkPumpPower(pumpIndex)     # Retrieves the constant power rating of the pumps given their indices
       
        See also getLinkPumpHCurve, getLinkPumpECurve, getLinkPumpECost,
                 getLinkPumpEPat, getLinkPumpPatternIndex, getLinkPumpPatternNameID.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_POWER, *argv)
    
    def getLinkPumpState(self, *argv):
        """ Retrieves the current computed pump state (read only) (see @ref EN_PumpStateType). (EPANET Version 2.2)
        same as status: open, active, closed
        Using step-by-step hydraulic analysis,
       
        Example 1:
          d.getLinkPumpState()        # Retrieves the current computed pump state for all links
       
        Example 2:
          d.getLinkPumpState(1)       # Retrieves the current computed pump state for the first link
       
        For more, you can check examples 3 & 4 of getLinkFlows function
       
        See also getLinkFlows, getLinkHeadloss, getLinkStatus,
                 getLinkSettings, getLinkEnergy, getLinkPumpEfficiency.
        """
        return self.get_pump_link_info(self.ToolkitConstants.EN_PUMP_STATE, *argv) 
       
    def getLinkPumpType(self):
        """ Retrieves the type of a pump. (EPANET Version 2.1)
       
        Example:
          d.getLinkPumpType()
       
        See also getLinkPumpTypeCode, getLinkPumpPower.
        """
        v = self.getLinkPumpTypeCode()
        pType = []
        for i in v:
            pType.append(self.TYPEPUMP[i])
        return pType
      
    def getLinkPumpTypeCode(self):
        """ Retrieves the code of type of a pump. (EPANET Version 2.1)
       
        Type of pump codes:
          0 = Constant horsepower
          1 = Power function
          2 = User-defined custom curve
       
        Example 1:
          d.getLinkPumpTypeCode()         #  Retrieves the all the  pumps type code
       
        Example 2:
          d.getLinkPumpTypeCode()[0]      #  Retrieves the first pump type code
        See also getLinkPumpType, getLinkPumpPower.
        """
        pumpCnt = self.getLinkPumpCount()
        j = 0
        value = [1] * pumpCnt
        if pumpCnt != 0:
            for i in self.getLinkPumpIndex():
                value[j] = self.d.ENgetpumptype(i)
                j += 1
        return value
        
    def getLinkValveCount(self):
        """ Retrieves the number of valves.
       
        Example:
          d.getLinkValveCount()
       
        See also getLinkPumpCount, getLinkCount.
        """
        try:
            linkTypes = self.getLinkTypeIndex()
            pipepump = linkTypes.count(self.ToolkitConstants.EN_CVPIPE) + linkTypes.count(self.ToolkitConstants.EN_PIPE) + linkTypes.count(self.ToolkitConstants.EN_PUMP)
            value = self.getLinkCount() - pipepump
        except Exception as Errcode:
            return self.getError(Errcode) 
            return
        return value
           
# Rewrite EXAMPLE 3 AND 4 USE FUNCTIONS THAT ARE NOT YET IMPLEMENTED
    def getLinkFlows(self, *argv):
        """ Retrieves the current computed flow rate (read only).
        Using step-by-step hydraulic analysis
       
        Example 1:
           d.getLinkFlows()        # Retrieves the current computed flow rate for all links
       
        Example 2:
           d.getLinkFlows(1)       # Retrieves the current computed flow rate for the first link
       
        Example 3:
             Hydraulic analysis step-by-step
           d.openHydraulicAnalysis() 
           d.initializeHydraulicAnalysis 
           tstep=1 P=[] T_H=[] D=[] H=[] F=[] S=[] 
           while (tstep>0)
              t= d.runHydraulicAnalysis 
              P = [P  d.getNodePressure] 
              D = [D  d.getNodeActualDemand] 
              H = [H  d.getNode HydraulicHead] 
              S = [S  d.getLinkStatus] 
              F = [F  d.getLinkFlows] 
              T_H = [T_H  t] 
              tstep=d.nextHydraulicAnalysisStep 
           end
           d.closeHydraulicAnalysis 
       
        Example 4: 
             Hydraulic and Quality analysis step-by-step
           d.openHydraulicAnalysis 
           d.openQualityAnalysis 
           d.initializeHydraulicAnalysis(0) 
           d.initializeQualityAnalysis(d.ToolkitConstants.EN_NOSAVE) 
        
           tstep = 1 
           T = []  P = []  F = []  QN = []  QL = [] 
           while (tstep>0)
              t  = d.runHydraulicAnalysis 
              qt = d.runQualityAnalysis 
        
              P  = [P  d.getNodePressure] 
              F  = [F  d.getLinkFlows] 
        
              QN = [QN  d.getNodeActualQuality] 
              QL = [QL  d.getLinkActualQuality] 
              T  = [T  t] 
        
              tstep = d.nextHydraulicAnalysisStep 
              qtstep = d.nextQualityAnalysisStep 
           end
           d.closeQualityAnalysis 
           d.closeHydraulicAnalysis 
       
        See also getLinkVelocity, getLinkHeadloss, getLinkStatus,
                 getLinkPumpState, getLinkSettings, getLinkEnergy,
                 getLinkActualQuality, getLinkPumpEfficiency.
        """
        return self.get_link_info(self.ToolkitConstants.EN_FLOW, *argv)  
    
    def getLinkVelocity(self, *argv):
        """ Retrieves the current computed flow velocity (read only).
       
        Using step-by-step hydraulic analysis
       
        Example 1:
          d.getLinkVelocity()        # Retrieves the current computed flow velocity for all links
       
        Example 2:
          d.getLinkVelocity(1)       # Retrieves the current computed flow velocity for the first link
       
        For more, you can check examples 3 & 4 of getLinkFlows function
       
        See also getLinkFlows, getLinkHeadloss, getLinkStatus,
                 getLinkPumpState, getLinkSettings, getLinkActualQuality.
        """
        return self.get_link_info(self.ToolkitConstants.EN_VELOCITY, *argv) 
    
    def getLinkVertices(self, *argv):
        """ Retrieves the coordinate's of a vertex point assigned to a link.
       
          The example is based on d=epanet('Net1.inp') 
       
        Example:
          d = epanet('Net1.inp') 
          linkID = '10' 
          x = [22, 24, 28] 
          y = [69, 68, 69] 
          d.setLinkVertices(linkID, x, y)
          linkID = '112' 
          x = [10, 24, 18] 
          y = [49, 58, 60] 
          d.setLinkVertices(linkID, x, y)
          d.getLinkVertices{1}
          d.getLinkVertices{2}
       
        See also getLinkVertices, getLinkVerticesCount, getBinLinkVertices,
                 getBinLinkVerticesCount.
        """
        try:
            if len(argv) > 0:
                if type(argv[0]) == str:
                    indices = self.getLinkIndex(argv[0])
                else:
                    indices = argv[0]
            else:
                indices = self.getLinkIndex()
            if type(indices) == int:
                indices = [indices]
            x_data = {}
            y_data = {}
            xy = []
            for i in indices:
                if self.getLinkVerticesCount(i) == [0]:
                    x_data[i] = []
                    y_data[i] = []
                    continue
                x_mat = []
                y_mat = []
                for j in self.getLinkVerticesCount(i):
                    xy = self.d.ENgetvertex(i, j)
                    x_mat.append(xy[0])
                    y_mat.append(xy[1])
                x_data[i] = x_mat
                y_data[i] = y_mat
        except Exception as Errcode:
            return self.getError(Errcode) 
        return {'x':x_data, 'y':y_data}
            
    def getLinkVerticesCount(self, *argv):
        """ Retrieves the number of internal vertex points assigned to a link.
         
          Example 1:
            d = epanet('anytown.inp')
            d.getLinkVerticesCount()          # Retrieves the vertices per link 
         
          Example 2:
            link_id = '2'
            d.getLinkVerticesCount(link_id)   # Retrieves the vertices of link '2' 
         
          Example 3:
            link_index = 31
            d.getLinkVerticesCount(link_index)    # Retrieves the vertices of link 31
         
          See also getLinkVertices, setLinkVertices, getBinLinkVertices,
                   getBinLinkVerticesCount.
        """
        try:
            if len(argv) > 0:
                if type(argv[0]) == str:
                    indices = self.getLinkIndex(argv[0])
                else:
                    indices = argv[0]
            else:
                indices = self.getLinkIndex()
            value = []
            if type(indices) == int:
                indices = [indices]
            for i in indices:
                value.append(self.d.ENgetvertexcount(i))
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value    
    
    def getLinkHeadloss(self, *argv):
        """ Retrieves the current computed head loss (read only).
         
          Using step-by-step hydraulic analysis,
         
          Example 1:
            d.getLinkHeadloss()      # Retrieves the current computed head loss for all links
         
          Example 2:
            d.getLinkHeadloss(1)     # Retrieves the current computed head loss for the first link
         
          For more, you can check examples 3 & 4 of getLinkFlows function
         
          See also getLinkFlows, getLinkVelocity, getLinkStatus, 
                   getLinkPumpState, getLinkSettings, getLinkActualQuality.\
        """
        return self.get_link_info(self.ToolkitConstants.EN_HEADLOSS, *argv) 

    def getLinkStatus(self, *argv):
        """ Retrieves the current link status (see @ref EN_LinkStatusType) (0 = closed, 1 = open).
         
          Using step-by-step hydraulic analysis,
         
          Example 1: 
            d.getLinkStatus()        # Retrieves the current link status for all links
         
          Example 2:
            d.getLinkStatus(1)       # Retrieves the current link status for the first link
         
          For more, you can check examples 3 & 4 of getLinkFlows function
         
          See also getLinkFlows, getLinkVelocity, getLinkHeadloss,
                   getLinkPumpState, getLinkSettings.
        """
        return self.get_link_info(self.ToolkitConstants.EN_STATUS, *argv) 

    def getLinkSettings(self, *argv):
        """ Retrieves the current computed value of all link roughness for pipes 
          or actual speed for pumps or actual setting for valves.
         
          Using step-by-step hydraulic analysis,
         
          Example 1:
            d.getLinkSettings()      # Retrieves the current values of settings for all links
         
          Example 2:
            d.getLinkSettings(1)     # Retrieves the current value of setting for the first link
         
          For more, you can check examples 3 & 4 of getLinkFlows function
         
          See also getLinkFlows, getLinkVelocity, getLinkHeadloss,
                   getLinkStatus, getLinkPumpState, getLinkEnergy.
        """
        return self.get_link_info(self.ToolkitConstants.EN_SETTING, *argv) 

    def getLinkEnergy(self, *argv):
        """ Retrieves the current computed pump energy usage (read only).
         
          Using step-by-step hydraulic analysis,
         
          Example 1:
            d.getLinkEnergy()        # Retrieves the current computed pump energy usage for all links
         
          Example 2:
            d.getLinkEnergy(1)       # Retrieves the current computed pump energy usage for the first link
         
          For more, you can check examples 3 & 4 of getLinkFlows function
         
          See also getLinkFlows, getLinkVelocity, getLinkHeadloss,
                   getLinkStatus, getLinkPumpState, getLinkPumpEfficiency.
        """
        return self.get_link_info(self.ToolkitConstants.EN_ENERGY, *argv)     
    
    def getLinkActualQuality(self, *argv):
        """ Retrieves the current computed link quality (read only). (EPANET Version 2.2)
         
          Example 1:
            d.getLinkActualQuality()        # Retrieves the current computed link quality for all links
         
          Example 2:
            d.getLinkActualQuality(1)       # Retrieves the current computed link quality for the first link
         
          Example 3: 
            check examples/EX14_hydraulic_and_quality_analysis.m
         
          See also getLinkFlows, getLinkStatus, getLinkPumpState,
                   getLinkSettings, getLinkPumpEfficiency.
        """
        return self.get_link_info(self.ToolkitConstants.EN_LINKQUAL, *argv)      
    
    def getLinkIndex(self, *argv):
        """ Retrieves the indices of all links, or the indices of an ID set of links.
         
          Example 1:
            d.getLinkIndex()                # Retrieves the indices of all links
         
          Example 2:
            linkID = d.getLinkNameID()[0]
            d.getLinkIndex(linkID)          # Retrieves the index of the 1st link given it's ID
         
          Example 3:
            linkID = d.getLinkNameID()[0:3]
            d.getLinkIndex(linkID)          # Retrieves the indices of the first 3 links given their ID
         
          See also getLinkNameID, getLinkPipeIndex, getNodeIndex.
        """
        try:
            values = [] 
            if len(argv) > 0:
                index = argv[0]
  # Check for int input (issue with last exception)
  # if not isinstance(index[0], str):
  #     warnings.warn('Name ID is needed as input (TYPE: string)')
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetlinkindex(i))
                else:
                    values = self.d.ENgetlinkindex(index) 
            else:
                for i in range(self.getLinkCount()):
                    values.append(i+1)
        except Exception as Errcode:
            return self.getError(Errcode)
        return values
    
    def getLinkNodesIndex(self, *argv):
        """ Retrieves the indexes of the from/to nodes of all links.
       
        Example 1:
          d.getLinkNodesIndex()
       
        Example 2:
          d.getLinkNodesIndex(2)    # Link index
       
        See also getNodesConnectingLinksID.
        """
        indices = self.getlinkIndices(*argv)
        value = []
        for i in indices:
            value.append(self.d.ENgetlinknodes(i))
        return value
    
    def getLinkPipeIndex(self, *argv):
        """ Retrieves the pipe indices.
         
          Example:
            d.getLinkPipeIndex()
         
          See also getLinkIndex, getLinkPumpIndex.
        """
        tmpLinkTypes = self.getLinkType() 
        value = []
        for i in range(len(tmpLinkTypes)):
            if tmpLinkTypes[i] == 'PIPE':
                value.append(i+1)
        return value
    
    def getLinkPipeNameID(self):
        """ Retrieves the pipe ID.
         
          Example 1:
            d.getLinkPipeNameID()         # Retrieves the ID's of all pipes
         
          Example 2:
            d.getLinkPipeNameID()[0]      # Retrieves the ID of the 1st pipe
         
          Example 3:
            d.getLinkPipeNameID()[0:3]    # Retrieves the ID of the first 3 pipes
         
          See also getLinkNameID, getLinkPumpNameID, getNodeNameID.
        """
        try:
            value = self.getLinkNameID(self.getLinkPipeIndex()) 
        except Exception as Errcode:
           return self.getError(Errcode)
        return value  
    
    def getLinkPumpIndex(self, *argv):
        """ Retrieves the pump indices.
         
          Example 1:
            d.getLinkPumpIndex()          # Retrieves the indices of all pumps
         
          Example 2:
            d.getLinkPumpIndex(1)         # Retrieves the index of the 1st pump
         
          Example 3:
            d.getLinkPumpIndex([1,2])     # Retrieves the indices of the first 2 pumps
         
          See also getLinkIndex, getLinkPipeIndex, getLinkValveIndex.
        """
        tmpLinkTypes = self.getLinkTypeIndex() 
        value = [i for i, x in enumerate(tmpLinkTypes) if x == self.ToolkitConstants.EN_PUMP]
        if (len(value) > 0) and (len(argv) > 0):
            index = argv[0]
            try:
                if type(index) is list:
                    pIndices = []
                    for i in index:
                        pIndices.append(value[i-1] + 1)
                    return pIndices
                else:
                    return value[index-1] + 1
            except:
                raise Epanet_Except('Some pump indices do not exist')
        else:
            pIndices = value
            return [i+1 for i in pIndices]
            
    def getLinkPumpNameID(self):
        """ Retrieves the pump ID.
         
          Example 1:
            d.getLinkPumpNameID()          # Retrieves the ID's of all pumps
         
          Example 2:
            d.getLinkPumpNameID()[0]       # Retrieves the ID of the 1st pump
         
          Example 3:
            d = epanet('Net3_trace.inp')
            d.getLinkPumpNameID()[0:2]     # Retrieves the ID of the first 2 pumps
         
          See also getLinkNameID, getLinkPipeNameID, getNodeNameID.
        """
        try:
            value = self.getLinkNameID(self.getLinkPumpIndex()) 
        except Exception as Errcode:
           return self.getError(Errcode)
        return value    
           
    def getLinkValveIndex(self):
        """ Retrieves the valve indices.
         
          Example:
            d.getLinkValveIndex()
         
          See also getLinkIndex, getLinkPipeIndex, getLinkPumpIndex.
        """
        tmpLinkTypes = self.getLinkType() 
        value = []
        for i in range(len(tmpLinkTypes)):
            if tmpLinkTypes[i] != 'PIPE' and tmpLinkTypes[i] != 'PUMP' :
                value.append(i+1)
        return value
    
    def getLinkValveNameID(self):
        """ Retrieves the valve ID.
         
          Example 1:
            d.getLinkValveNameID()          # Retrieves the ID's of all valves
         
          Example 2:
            d.getLinkValveNameID()[0]       # Retrieves the ID of the 1st valve
         
          Example 3:
            d.getLinkValveNameID()[0:3]     # Retrieves the ID of the first 3 valves
         
          See also getLinkNameID, getLinkPumpNameID, getNodeNameID.
        """
        try:
            value = self.getLinkNameID(self.getLinkValveIndex()) 
        except Exception as Errcode:
             return self.getError(Errcode) 
        return value
    
    def getNodeCount(self):
        """ Retrieves the number of nodes.
         
          Example:
            d.getNodeCount()
         
          See also getNodeIndex, getLinkCount.
        """
        try:
            value = self._d.ENgetcount(self.ToolkitConstants.EN_NODECOUNT)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value
    
    def getNodeActualDemand(self, *argv):
        """ Retrieves the computed value of all node actual demands.
         
            Example 1:
              d.getNodeActualDemand()           # Retrieves the computed value of all node actual demands
             
            Example 2:
              d.getNodeActualDemand(1)          # Retrieves the computed value of the first node actual demand
             
            Example 3:
              d.getNodeActualDemand([1,2,3])    # Retrieves the computed value of the first 3 nodes actual demand
         
         
            See also getNodeActualDemandSensingNodes, getNode HydraulicHead, getNodePressure,
                     getNodeActualQuality, getNodeMassFlowRate, getNodeActualQualitySensingNodes.
        """
        return self.get_node_info(self.ToolkitConstants.EN_DEMAND, *argv)
    
    def getNodeActualQuality(self, *argv):
         """ Retrieves the computed values of the actual quality for all nodes.
         
          Example 1:
            d.getNodeActualQuality()        # Retrieves the computed values of the actual quality for all nodes
         
          Example 2:
            d.getNodeActualQuality(1)       # Retrieves the computed value of the actual quality for the first node
         
          See also getNodeActualDemand, getNodeActualDemandSensingNodes, getNodePressure,
                   getNode HydraulicHead, getNodeMassFlowRate, getNodeActualQualitySensingNodes.
         """
         return self.get_node_info(self.ToolkitConstants.EN_QUALITY, *argv) 
    
    def getNodeBaseDemands(self, *argv):
        """ Retrieves the value of all node base demands.
       
        Example 1:
          d.getNodeBaseDemands()
          d.getNodeBaseDemands()[0]      #  Get categories 1
       
        Example 2:
          d.getNodeBaseDemands(2)        # Get node base demand with categories for specific node index
       
        See also setNodeBaseDemands, getNodeDemandCategoriesNumber,
                 getNodeDemandPatternIndex, getNodeDemandPatternNameID.
        """
        indices = self.getNodeIndices(*argv)
        numdemands = self.getNodeDemandCategoriesNumber(indices)
        value =  dict.fromkeys(range(max(numdemands)), [])
        val = np.zeros((max(numdemands), len(indices)))
        j = 1
        for i in indices:
            v = 0
            for u in range(numdemands[j-1]):
                val[v][j-1] = self.d.ENgetbasedemand(i, u+1)
                v+=1
            j+=1
        for i in range(max(numdemands)):
            value[i] = list(val[i]) 
        return value  
    
    def getNodeCoordinates(self, *argv):
        if (self.getLinkCount() is None): # Does it stand ????????
            raise Epanet_Except('Not enough network coordinates.')
        # GET VERTICES 
        vertices = self.getLinkVertices()
        # SET X Y node coordinates 
        indices = self.getNodeIndices(*argv)
        vx, vy = [], []
        try:    
            for i in indices:
                 vx.append(self.d.ENgetcoord(i)[0])
                 vy.append(self.d.ENgetcoord(i)[1])
        except Exception as Errcode:
            return self.getError(Errcode)
        if len(argv) == 0:
            vxx, vyy = {}, {}
            for i in indices:
                vxx[i] = vx[i-1]
                vyy[i] = vy[i-1]
            return {'x':vxx, 'y':vyy, 'x_vert':vertices['x'], 'y_vert':vertices['y']}
        else:
            xVal, yVal = {}, {}
            j = 1
            for i in indices:
                xVal[i] = vx[j-1]
                yVal[i] = vy[j-1]
                j+=1
            return {'x':xVal, 'y':yVal}
                    
    def getNodeDemandCategoriesNumber(self, *argv):
        """ Retrieves the value of all node base demands categorie number. (EPANET Version 2.1)
       
        Example 1:
       	d.getNodeDemandCategoriesNumber()               # Retrieves the value of all node base demands categorie number
       
        Example 2:
       	d.getNodeDemandCategoriesNumber(1)              # Retrieves the value of the first node base demand categorie number
        
        Example 3:
       	d.getNodeDemandCategoriesNumber([1,2,3,4])      # Retrieves the value of the first 4 nodes base demand categorie number   
       
        See also getNodeBaseDemands, getNodeDemandPatternIndex,
                  getNodeDemandPatternNameID.
        """
        try:
             value = []
             if len(argv) > 0:
                 index = argv[0]
                 if type(index) is list:
                     for i in index:
                         value.append(self.d.ENgetnumdemands(i)) 
                 else:
                     value = self.d.ENgetnumdemands(index) 
             else:
                 for i in range(self.getNodeCount()):
                     value.append(self.d.ENgetnumdemands(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return value

    def getNodeDemandDeficit(self, *argv):
        """  Retrieves the amount that full demand is reduced under PDA. (EPANET Version 2.2)
       
        The example is based on d=epanet('NET1.inp') 
       
        Example:
          d.setDemandModel('PDA', 0, 0.1, 0.5)      # Sets a type of demand model and its parameters
          d.getComputedHydraulicTimeSeries()        # Computes hydraulic simulation and retrieve all time-series
          d.getNodeDemandDeficit                    # Retrieves the amount that full demand is reduced under PDA
       
        See also setDemandModel, getComputedHydraulicTimeSeries,
                 getNodeActualDemand, getNodeActualDemandSensingNodes.
         """
        return self.get_node_info(self.ToolkitConstants.EN_DEMANDDEFICIT, *argv)
    
    @property
    def getNodeDemandPatternIndex(self):
        """ Retrieves the value of all node base demands pattern index. (EPANET Version 2.1)
       
        Example:
          d.getNodeDemandPatternIndex()
          d.getNodeDemandPatternIndex()[1]
       
        See also getNodeBaseDemands, getNodeDemandCategoriesNumber,
                 getNodeDemandPatternNameID,
                 setNodeDemandPatternIndex.
        """
        numdemands = self.getNodeDemandCategoriesNumber()
        value =  dict.fromkeys(range(max(numdemands)), [])
        val = np.zeros((max(numdemands), self.getNodeCount()))
        for i in self.getNodeIndex():
            v = 0
            for u in range(numdemands[i-1]):
                val[v][i-1] = self.d.ENgetdemandpattern(i, u+1)
                v+=1      
        for i in self.getNodeReservoirIndex():            
            val[v][i-1] = self.d.ENgetnodevalue(i, self.ToolkitConstants.EN_PATTERN)
        for i in range(max(numdemands)):
            value[i] = list(val[i]) 
        return value
    
    def getNodeElevations(self, *argv):
        """ Retrieves the value of all node elevations.
        Example 1:
          d.getNodeElevations()             # Retrieves the value of all node elevations
       
        Example 2:
          d.getNodeElevations(1)            # Retrieves the value of the first node elevation
       
        Example 3:
          d.getNodeElevations([4, 5, 6])    # Retrieves the value of the 5th to 7th node elevations
       
        See also setNodeElevations, getNodesInfo, getNodeNameID,
                   getNodeType, getNodeEmitterCoeff, getNodeInitialQuality. 
        """
        return self.get_node_info(self.ToolkitConstants.EN_ELEVATION, *argv)
    
    def getNodeEmitterCoeff(self, *argv):
        """ Retrieves the value of all node emmitter coefficients.
       
        Example 1:
          d.getNodeEmitterCoeff()         # Retrieves the value of all node emmitter coefficients
       
        Example 2:
          d.getNodeEmitterCoeff(1)        # Retrieves the value of the first node emmitter coefficient
       
        See also setNodeEmitterCoeff, getNodesInfo, getNodeElevations.
        """
        return self.get_node_info(self.ToolkitConstants.EN_EMITTER, *argv)
    
    def getNodeHydraulicHead(self, *argv):
        """ Retrieves the computed values of all node hydraulic heads.
       
        Example 1:
          d.getNodeHydraulicHead()        # Retrieves the computed value of all node hydraulic heads
       
        Example 2:
          d.getNodeHydraulicHead(1)       # Retrieves the computed value of the first node hydraulic head
       
        For more, you can type `help getNodePressure` and check examples 3 & 4.
       
        See also getNodeActualDemand, getNodeActualDemandSensingNodes, getNodePressure,
                 getNodeActualQuality, getNodeMassFlowRate, getNodeActualQualitySensingNodes.
        """
        return self.get_node_info(self.ToolkitConstants.EN_HEAD, *argv)
     
    def getNodeIndex(self, *argv):
        """ Retrieves the indices of all nodes or some nodes with a specified ID.
       
        Example 1:
          d.getNodeIndex()              # Retrieves the indices of all nodes
       
        Example 2:
          nameID = d.getNodeNameID(1)
          d.getNodeIndex(nameID)        # Retrieves the node index given the ID label of the 1st node
       
        See also getNodeNameID, getNodeReservoirIndex, getNodeJunctionIndex,
                 getNodeType, getNodeTypeIndex, getNodesInfo.
        """
        try:
            values = [] 
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetnodeindex(i))
                else:
                    values = self.d.ENgetnodeindex(index) 
            else:
                for i in range(self.getNodeCount()):
                    values.append(i+1)
        except Exception as Errcode:
            return self.getError(Errcode)
        return values
    
    def getNodeInitialQuality(self, *argv):
        """ Retrieves the value of all node initial quality.
       
        Example 1:
          d.getNodeInitialQuality()          # Retrieves the value of all node initial quality
       
        Example 2:
          d.getNodeInitialQuality(1)         # Retrieves the value of the first node initial quality
       
        See also setNodeInitialQuality, getNodesInfo, getNodeSourceQuality.
        """
        return self.get_node_info(self.ToolkitConstants.EN_INITQUAL, *argv)
    
    def getNodeJunctionCount(self):
        """ Retrieves the number of junction nodes.
         
          Example:
            d.getNodeJunctionCount()
         
          See also getNodeTankCount, getNodeCount.
        """
        return self.getNodeTypeIndex().count(self.ToolkitConstants.EN_JUNCTION)
    
    def getNodeJunctionIndex(self, *argv):
        """Retrieves the indices of junctions.
         
          Example 1:
            d.getNodeJunctionIndex()          # Retrieves the indices of all junctions
            
          Example 2:
            d.getNodeJunctionIndex([1,2])     # Retrieves the indices of the first 2 junctions
         
          See also getNodeNameID, getNodeIndex, getNodeReservoirIndex,
                   getNodeType, getNodeTypeIndex, getNodesInfo.
        """           
        tmpNodeTypes = self.getNodeTypeIndex()
        value = [i for i, x in enumerate(tmpNodeTypes) if x == self.ToolkitConstants.EN_JUNCTION]
        if (len(value) > 0) and (len(argv) > 0):
            index = argv[0]
            try:
                if type(index) is list:
                    jIndices = []
                    for i in index:
                        jIndices.append(value[i-1] + 1)
                    return jIndices
                else:
                    return value[index-1] + 1
            except:
                raise Epanet_Except('Some Junction indices do not exist')
        else:
            jIndices = value
            return [i+1 for i in jIndices]
    
    def getNodePatternIndex(self, *argv):
        """  Retrieves the value of all node demand pattern indices.
         
          Example 1:
            d.getNodePatternIndex()        #  Retrieves the value of all node demand pattern indices
            
          Example 2:
            d.getNodePatternIndex(1)       #  Retrieves the value of the first node demand pattern index
         
          See also getNodeBaseDemands, getNodeDemandCategoriesNumber,
                   getNodeDemandPatternIndex, getNodeDemandPatternNameID.
        """
        return self.get_node_info(self.ToolkitConstants.EN_PATTERN, *argv) 
    
    def getNodePressure(self, *argv):
        """ Retrieves the computed values of all node pressures.
       
        Example 1:
           d.getNodePressure()          # Retrieves the computed values of all node pressures
       
        Example 2:
           d.getNodePressure(1)         # Retrieves the computed value of the first node pressure
       
        Example 3:
             Hydraulic analysis step-by-step
           d.openHydraulicAnalysis
           d.initializeHydraulicAnalysis() 
           tstep,P , T_H, D, H, F, S, = 1, [], [], [], [] ,[], [] 
           while (tstep>0):
              t = d.runHydraulicAnalysis 
              P.append(d.getNodePressure())
              D.append(d.getNodeActualDemand()) 
              H.append(d.getNodeHydaulicHead()) 
              S.append(d.getLinkStatus()) 
              F.append(d.getLinkFlows()) 
              T_H.append(t)
              tstep=d.nextHydraulicAnalysisStep 
           end
           d.closeHydraulicAnalysis 
       
        Example 4:
            Hydraulic and Quality analysis step-by-step
           d.openHydraulicAnalysis 
           d.openQualityAnalysis 
           d.initializeHydraulicAnalysis(0) 
           d.initializeQualityAnalysis(d.ToolkitConstants.EN_NOSAVE) 
       
           tstep = 1 
           T, P, F, QN, QL = [], [], [], [], [] 
           while (tstep>0):
              t  = d.runHydraulicAnalysis 
              qt = d.runQualityAnalysis 
       
              P  = [P  d.getNodePressure] 
              F  = [F  d.getLinkFlows] 
       
              QN = [QN  d.getNodeActualQuality] 
              QL = [QL  d.getLinkActualQuality] 
              T  = [T  t] 
       
              tstep = d.nextHydraulicAnalysisStep 
              qtstep = d.nextQualityAnalysisStep 
           end
           d.closeQualityAnalysis 
           d.closeHydraulicAnalysis 
       
        See also getNodeActualDemand, getNodeActualDemandSensingNodes, getNodeHydaulicHead
                 getNodeActualQuality, getNodeMassFlowRate, getNodeActualQualitySensingNodes.
        """
        return self.get_node_info(self.ToolkitConstants.EN_PRESSURE, *argv)

    def getNodeReservoirIndex(self, *argv):
        """ Retrieves the indices of reservoirs.
       
        Example 1:
          d.getNodeReservoirIndex()           # Retrieves the indices of all reservoirs.
       
        Example 2:
          d.getNodeReservoirIndex([1,2,3])    # Retrieves the indices of the first 3 reservoirs.
       
        See also getNodeNameID, getNodeIndex, getNodeJunctionIndex,
                 getNodeType, getNodeTypeIndex, getNodesInfo.
        """
        tmpNodeTypes = self.getNodeTypeIndex()
        value = [i for i, x in enumerate(tmpNodeTypes) if x == self.ToolkitConstants.EN_RESERVOIR]
        if (len(value) > 0) and (len(argv) > 0):
            index = argv[0]
            try:
                if type(index) is list:
                    rIndices = []
                    for i in index:
                        rIndices.append(value[i-1] + 1)
                    return rIndices
                else:
                    return value[index-1] + 1
            except:
                raise Epanet_Except('Some Reservoirs indices do not exist')
        else:
            rIndices = value
            return [i+1 for i in rIndices]
        
    def getNodesConnectingLinksID(self, *argv):
        """ Retrieves the id of the from/to nodes of all links.
       
        Example 1:
          d.getNodesConnectingLinksID()            # Retrieves the id of the from/to nodes of all links
       
        Example 2:
          linkIndex = 1;
          d.getNodesConnectingLinksID(1)           # Retrieves the id of the from/to nodes of the 1st link
       
        See also getLinkNodesIndex.
        """
        indices = self.getlinkIndices(*argv)
        values  = self.getLinkNodesIndex(indices)
        connVals = []
        for i in range(len(values)):
            connVals.append(self.getNodeNameID((values[i])))
        return connVals

    def getNodesConnectingLinksIndex(self):
        """ Retrieves the indexes of the from/to nodes of all links.
        Duplicate function with getLinkNodesIndex for new version.
           
        Example:
          d.getNodesConnectingLinksIndex()
           
        See also getLinkNodesIndex, getNodesConnectingLinksID.
        """
        return self.getLinkNodesIndex()
        
    def getNodeNameID(self, *argv):
        """ Retrieves the ID label of all nodes or some nodes with a specified index.
       
        Example 1:
          d.getNodeNameID()                   # Retrieves the ID label of all nodes
      
        Example 2:
          d.getNodeNameID(1)                  # Retrieves the ID label of the first node
       
        Example 3:
          junctionIndex = d.getNodeJunctionIndex()
          d.getNodeNameID(junctionIndex)       # Retrieves the ID labels of all junctions give their indices
       
        See also getNodeReservoirNameID, getNodeJunctionNameID,
                 getNodeIndex, getNodeType, getNodesInfo.
        """
        try:
            values = [] 
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetnodeid(i))
                else:
                    values = self.d.ENgetnodeid(index) 
            else:
                for i in range(self.getNodeCount()):
                    values.append(self.d.ENgetnodeid(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return values
    
    def getNodeReservoirCount(self):
        """ Retrieves the number of Reservoirs.
       
        Example:
          d.getNodeReservoirCount()
       
        See also getNodeTankCount, getNodeCount.
        """
        return self.getNodeTypeIndex().count(self.ToolkitConstants.EN_RESERVOIR)
    
    def getNodeSourcePatternIndex(self, *argv):
        """ Retrieves the value of all node source pattern index.
       
        Example 1:
          d.getNodeSourcePatternIndex()       #  Retrieves the value of all node source pattern index
       
        Example 2:
          d.getNodeSourcePatternIndex(1)      #  Retrieves the value of the first node source pattern index
       
        See also setNodeSourcePatternIndex, getNodeSourceQuality,
                 getNodeSourceTypeIndex, getNodeSourceType.
        """
        indices = self.getNodeIndices(*argv)
        value = []
        j=1
        for i in indices:
            try:
                value.append(self.d.ENgetnodevalue(i, self.ToolkitConstants.EN_SOURCEPAT))
            except Exception as Errcode:   
                if Errcode.args[0][13:16] == '203':
                    return self.getError(Errcode)
                value.append(None)
            j=j+1
        return value
    
    def getNodeSourceQuality(self, *argv):
        """ Retrieves the value of all node source quality.
       
        Example 1:
          d.getNodeSourceQuality()         # Retrieves the value of all node source quality
       
        Example 2:
          d.getNodeSourceQuality(1)        # Retrieves the value of the first node source quality
       
        See also setNodeSourceQuality, getNodeInitialQuality, getNodeSourcePatternIndex,
                 getNodeSourceTypeIndex, getNodeSourceType.
        """
        indices = self.getNodeIndices(*argv)
        value = []
        j=1
        for i in indices:
            try:
                value.append(self.d.ENgetnodevalue(i, self.ToolkitConstants.EN_SOURCEQUAL))
            except Exception as Errcode:   
                if Errcode.args[0][13:16] == '203':
                    return self.getError(Errcode)
                value.append(0)
            j=j+1
        return value
    
    def getNodeSourceType(self, *argv):
        """ Retrieves the value of all node source type.
       
        Example 1:
          d.getNodeSourceType()        # Retrieves the value of all node source type
       
        Example 2:
          d.getNodeSourceType(1)       # Retrieves the value of the first node source type
       
        See also setNodeSourceType, getNodeSourceQuality,
                 getNodeSourcePatternIndex, getNodeSourceTypeIndex.
        """
        indices = self.getNodeIndices(*argv)
        value =  dict.fromkeys(range(len(indices)), [])
        j = 0
        for i in indices:
            try:
                temp = self.d.ENgetnodevalue(i, self.ToolkitConstants.EN_SOURCETYPE)
            except Exception as Errcode:   
                if Errcode.args[0][13:16] == '203':
                    return self.getError(Errcode)
                temp = None 
            if temp != None:
                    value[j+1] = self.TYPESOURCE[temp+1]
            else:
                    value[j+1] = []
            j+=1
        return value
        
    def getNodeSourceTypeIndex(self, *argv):
        """ Retrieves the value of all node source type index.
       
        Example 1:
          d.getNodeSourceTypeIndex()        # Retrieves the value of all node source type index
       
        Example 2:
          d.getNodeSourceTypeIndex(1)       # Retrieves the value of the first node source type index
       
        See also getNodeSourceQuality, getNodeSourcePatternIndex, getNodeSourceType.
        """
        indices = self.getNodeIndices(*argv)
        value = []
        j=1
        for i in indices:
            try:
                value.append(self.d.ENgetnodevalue(i, self.ToolkitConstants.EN_SOURCETYPE))
            except Exception as Errcode:   
                if Errcode.args[0][13:16] == '203':
                    return self.getError(Errcode)
                value.append(None)  
            j=j+1
        return value
    
    def getNodeTankBulkReactionCoeff(self, *argv):
        """ Retrieves the tank bulk rate coefficient.
       
        Example 1:
          d.getNodeTankBulkReactionCoeff()                 # Retrieves the bulk rate coefficient of all tanks
       
        Example 2:
          d.getNodeTankBulkReactionCoeff(1)                # Retrieves the bulk rate coefficient of the 1st tank
       
        Example 3:
          d.getNodeTankBulkReactionCoeff([1,2])            # Retrieves the bulk rate coefficient of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankBulkReactionCoeff(tankIndex)        # Retrieves the bulk rate coefficient of the tanks given their indices
       
        See also setNodeTankBulkReactionCoeff, getNodeTankData.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_TANK_KBULK, *argv)
        
    def getNodeTankCount(self):
        """ Retrieves the number of Tanks.
       
        Example:
          d.getNodeTankCount()
       
          See also getNodeReservoirCount, getNodeCount.
        """
        return self.getNodeTypeIndex().count(self.ToolkitConstants.EN_TANK) 
    
    def getNodeTankDiameter(self, *argv):
        """ Retrieves the tank diameters.
       
        Example 1:
          d.getNodeTankDiameter()                # Retrieves the diameters of all tanks
       
        Example 2:
          d.getNodeTankDiameter(1)               # Retrieves the diameter of the 1st tank
       
        Example 3:
          d.getNodeTankDiameter([1,2])           # Retrieves the diameters of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankDiameter(tankIndex)       # Retrieves the diameters of the tanks given their indices
       
        See also setNodeTankDiameter, getNodeTankBulkReactionCoeff, getNodeTankInitialLevel,
                 getNodeTankMixingModelType, getNodeTankVolume, getNodeTankNameID.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_TANKDIAM, *argv)
    
    def getNodeTankIndex(self, *argv):
        """ Retrieves the tank indices.
       
        Example 1:
          d.getNodeTankIndex()       # Retrieves the tank indices.
       
        Example 2:
          d.getNodeTankIndex(1)      # Retrieves the first tank index.
       
        See also getNodeTankCount, getNodeTankNameID.
        """
        tmpNodeTypes = self.getNodeTypeIndex() 
        value = [i for i, x in enumerate(tmpNodeTypes) if x == self.ToolkitConstants.EN_TANK]
        if (len(value) > 0) and (len(argv) > 0):
            index = argv[0]
            try:
                if type(index) is list:
                    tIndices = []
                    for i in index:
                        tIndices.append(value[i-1] + 1)
                    return tIndices
                else:
                    return value[index-1] + 1
            except:
                raise Epanet_Except('Some tank indices do not exist')
        else:
            tIndices = value
            return [i+1 for i in tIndices]
        
    def getNodeTankInitialLevel(self, *argv):
        """ Retrieves the value of all tank initial water levels.
       
        Example 1:
          d.getNodeTankInitialLevel()         # Retrieves the value of all tank initial water levels
       
        Example 2:
          d.getNodeTankInitialLevel(11)       # Retrieves the value of the eleventh node(tank) water level
       
        See also setNodeTankInitialLevel, getNodeTankInitialWaterVolume, getNodeTankVolume,
                 getNodeTankMaximumWaterLevel, getNodeTankMinimumWaterLevel.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_TANKLEVEL, *argv) 
    
    def getNodeTankInitialWaterVolume(self, *argv):
        """ Retrieves the tank initial water volume.
       
        Example 1:
          d.getNodeTankInitialWaterVolume()                 #  Retrieves the initial water volume of all tanks
       
        Example 2:
          d.getNodeTankInitialWaterVolume(1)                #  Retrieves the initial water volume of the 1st tank
       
        Example 3:
          d.getNodeTankInitialWaterVolume([1,2])            #  Retrieves the initial water volume of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex 
          d.getNodeTankInitialWaterVolume(tankIndex)     Retrieves the initial water volume of the tanks given their indices
       
        See also getNodeTankInitialLevel,  getNodeTankVolume,
                     getNodeTankMaximumWaterVolume, getNodeTankMinimumWaterVolume.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_INITVOLUME, *argv) 
    
    def getNodeTankMaximumWaterLevel(self, *argv):
        """ Retrieves the tank maximum water level.
       
        Example 1:
          d.getNodeTankMaximumWaterLevel()                # Retrieves the maximum water level of all tanks
       
        Example 2:
          d.getNodeTankMaximumWaterLevel(1)               # Retrieves the maximum water level of the 1st tank
       
        Example 3:
          d.getNodeTankMaximumWaterLevel([1,2])           # Retrieves the maximum water level of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankMaximumWaterLevel(tankIndex)       # Retrieves the maximum water level of the tanks given their indices
       
        See also setNodeTankMaximumWaterLevel, getNodeTankMinimumWaterLevel, getNodeTankInitialLevel,
                 getNodeTankMaximumWaterVolume, getNodeTankMinimumWaterVolume, getNodeTankVolume.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MAXLEVEL, *argv) 
    
    def getNodeTankMaximumWaterVolume(self, *argv):
        """ Retrieves the tank maximum water volume. (EPANET Version 2.1)
       
        Example 1:
          d.getNodeTankMaximumWaterVolume                Retrieves the maximum water volume of all tanks
       
        Example 2:
          d.getNodeTankMaximumWaterVolume(1)             Retrieves the maximum water volume of the 1st tank
       
        Example 3:
          d.getNodeTankMaximumWaterVolume(1:2)           Retrieves the maximum water volume of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex;
          d.getNodeTankMaximumWaterVolume(tankIndex)     Retrieves the maximum water volume of the tanks given their indices
       
        See also getNodeTankMinimumWaterVolume, getNodeTankData.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MAXVOLUME, *argv) 
    
    def getNodeTankMinimumWaterLevel(self, *argv):
        """ Retrieves the tank minimum water level.
       
        Example 1:
          d.getNodeTankMinimumWaterLevel()                # Retrieves the minimum water level of all tanks
       
        Example 2:
          d.getNodeTankMinimumWaterLevel(1)               # Retrieves the minimum water level of the 1st tank
       
        Example 3:
          d.getNodeTankMinimumWaterLevel([1,2])           # Retrieves the minimum water level of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex;
          d.getNodeTankMinimumWaterLevel(tankIndex)       # Retrieves the minimum water level of the tanks given their indices
       
        See also setNodeTankMinimumWaterLevel, getNodeTankMaximumWaterLevel, getNodeTankInitialLevel,
                 getNodeTankMaximumWaterVolume, getNodeTankMinimumWaterVolume, getNodeTankVolume.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MINLEVEL, *argv) 
    
    def getNodeTankMinimumWaterVolume(self, *argv):
        """ Retrieves the tank minimum water volume.
       
        Example 1:
          d.getNodeTankMinimumWaterVolume()                # Retrieves the minimum water volume of all tanks
       
        Example 2:
          d.getNodeTankMinimumWaterVolume(1)               # Retrieves the minimum water volume of the 1st tank
       
        Example 3:
          d.getNodeTankMinimumWaterVolume([1,2])           # Retrieves the minimum water volume of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankMinimumWaterVolume(tankIndex)       # Retrieves the minimum water volume of the tanks given their indices
       
        See also setNodeTankMinimumWaterVolume, getNodeTankMaximumWaterVolume, getNodeTankInitialWaterVolume,
                 getNodeTankInitialLevel,  getNodeTankVolume, getNodeTankMixZoneVolume.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MINVOLUME, *argv)
    
    def getNodeTankMixingFraction(self, *argv):
        """ Retrieves the tank Fraction of total volume occupied by the inlet/outlet zone in a 2-compartment tank.
       
        Example 1:
           d.getNodeTankMixingFraction()                # Retrieves the mixing fraction of all tanks
       
        Example 2:
          d.getNodeTankMixingFraction(1)                # Retrieves the mixing fraction of the 1st tank
       
        Example 3:
          d.getNodeTankMixingFraction([1,2])            # Retrieves the mixing fraction of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankMixingFraction(tankIndex)        # Retrieves the mixing fraction of the tanks given their indices
       
        See also setNodeTankMixingFraction, getNodeTankData.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MIXFRACTION, *argv)
    
    def getNodeTankMixingModelCode(self, *argv):
        """ Retrieves the tank mixing model code.
       
        Code meaning:
          0 = Complete mix model (MIX1)
          1 = 2-compartment model (MIX2)
          2 = First in, first out model (FIFO)
          3 = Last in, first out model (LIFO)
       
        Example 1:
          d.getNodeTankMixingModelCode()                # Retrieves the mixing model code of all tanks
       
        Example 2:
          d.getNodeTankMixingModelCode(1)               # Retrieves the mixing model code of the 1st tank
       
        Example 3:
          d.getNodeTankMixingModelCode([1,2])           # Retrieves the mixing model code of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex;
          d.getNodeTankMixingModelCode(tankIndex)       # Retrieves the mixing model code of the tanks given their indices
       
        See also setNodeTankMixingModelType, getNodeTankMixingModelType, getNodeTankMixZoneVolume.
        """
        return self.get_node_tank_mixining_model(*argv)[0]
    
    def getNodeTankMixingModelType(self, *argv):
        """ Retrieves the tank mixing model type.
       
        Types of models that describe water quality mixing in storage tanks:
          MIX1 = Complete mix model
          MIX2 = 2-compartment model
          FIFO = First in, first out model
          LIFO = Last in, first out model
       
        Example 1:
          d.getNodeTankMixingModelType()               # Retrieves the mixing model type of all tanks
       
        Example 2:
          d.getNodeTankMixingModelType(1)              # Retrieves the mixing model type of the 1st tank
       
        Example 3:
          d.getNodeTankMixingModelType([1,2])          # Retrieves the mixing model type of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankMixingModelType(tankIndex)      # Retrieves the mixing model type of the tanks given their indices
       
        See also setNodeTankMixingModelType, getNodeTankMixingModelCode, getNodeTankMixZoneVolume
        """
        return self.get_node_tank_mixining_model(*argv)[1]

    def getNodeTankMixZoneVolume(self, *argv):
        """ Retrieves the tank mixing zone volume.
       
        Example 1:
          d.getNodeTankMixZoneVolume ()               # Retrieves the mixing zone volume of all tanks
       
        Example 2:
          d.getNodeTankMixZoneVolume(1)               # Retrieves the mixing zone volume of the 1st tank
       
        Example 3:
          d.getNodeTankMixZoneVolume([1,2])           # Retrieves the mixing zone volume of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex 
          d.getNodeTankMixZoneVolume(tankIndex)       # Retrieves the mixing zone volume of the tanks given their indices
       
        See also getNodeTankMixingModelCode, getNodeTankMixingModelType.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_MIXZONEVOL, *argv)
    
    def getNodeTankVolume(self, *argv):
        """ Retrieves the tank volume. (EPANET Version 2.1)
       
        Example 1:
          d.getNodeTankVolume()                  # Retrieves the volume of all tanks
       
        Example 2:
          d.getNodeTankVolume(1)                 # Retrieves the volume of the 1st tank
       
        Example 3:
          d.getNodeTankVolume([1,2])             # Retrieves the volume of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankVolume(tankIndex)         # Retrieves the volume of the tanks given their indices
       
        See also getNodeTankData.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_TANKVOLUME, *argv)

    
    def getNodeTankVolumeCurveIndex(self, *argv):
        """ Retrieves the tank volume curve index.
       
        Example 1:
          d.getNodeTankVolumeCurveIndex()                # Retrieves the volume curve index of all tanks
       
        Example 2:
          d.getNodeTankVolumeCurveIndex(1)               # Retrieves the volume curve index of the 1st tank
       
        Example 3:
          d.getNodeTankVolumeCurveIndex([1,2])           # Retrieves the volume curve index of the first 2 tanks
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.getNodeTankVolumeCurveIndex(tankIndex)       # Retrieves the volume curve index of the tanks given their indices
       
        See also getNodeTankVolume, getNodeTankMaximumWaterVolume, getNodeTankMinimumWaterVolume,
                 getNodeTankInitialWaterVolume, getNodeTankMixZoneVolume.
        """
        return self.get_tank_node_info(self.ToolkitConstants.EN_VOLCURVE, *argv)
    
    def getNodeType(self, *argv):
        """ Retrieves the node-type code for all nodes.
         
          Example 1: 
            d.getNodeType()          # Retrieves the node-type code for all nodes
         
          Example 2:
            d.getNodeType(1)         # Retrieves the node-type code for the first node
         
          Example 3:
            d.getNodeType([10,11])   # Retrieves the node-type code for the tenth and eleventh nodes
         
          See also getNodeNameID, getNodeIndex, getNodeTypeIndex, getNodesInfo.
        """
        try:
            nTypes = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        nTypes.append(self.TYPENODE[self.d.ENgetnodetype(i)])
                else:
                    nTypes = self.TYPENODE[self.d.ENgetnodetype(index)] 
            else:
                for i in range(self.getNodeCount()):
                    nTypes.append(self.TYPENODE[self.d.ENgetnodetype(i+1)])
        except Exception as Errcode:
            return self.getError(Errcode)
        return nTypes
    
    def getNodeTypeIndex(self, *argv):
        """ Retrieves the node-type code for all nodes.
       
        Example 1:
          d.getNodeTypeIndex()      # Retrieves the node-type code for all nodes
       
        Example 2:
          d.getNodeTypeIndex(1)     # Retrieves the node-type code for the first node
       
        See also getNodeNameID, getNodeIndex, getNodeType, getNodesInfo.
        """
        try:
            nTypes = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        nTypes.append(self.d.ENgetnodetype(i))
                else:
                    nTypes = self.d.ENgetnodetype(index) 
            else:
                for i in range(self.getNodeCount()):
                    nTypes.append(self.d.ENgetnodetype(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return nTypes
        
    def getPatternCount(self):
        """ Retrieves the number of patterns.
       
        Example:
          d.getPatternCount()
       
        See also getPatternIndex, getPattern.
        """
        try:
            value = self.d.ENgetcount(self.ToolkitConstants.EN_PATCOUNT)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value
    
    def getRuleCount(self):
        """ Retrieves the number of rules. (EPANET Version 2.2)
       
        Example:
          d.getRuleCount()
       
        See also getRules, getControlRulesCount.
        """
        try:
            value = self.d.ENgetcount(6)
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value
    
    @property
    def getTimeSimulationDuration(self):
        """ Retrieves the value of simulation duration.
       
        Example:
          d.getTimeSimulationDuration
       
        See also getTimePatternStep, getTimeHydraulicStep.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_DURATION)
    
    @property
    def getTimeHydraulicStep(self):
        """ Retrieves the value of the hydraulic time step.
       
        Example:
          d.getTimeHydraulicStep
       
        See also getTimeQualityStep, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_HYDSTEP)
    
    @property
    def getTimeQualityStep(self):
        """ Retrieves the value of the water quality time step.
       
        Example:
          d.getTimeQualityStep
       
        See also getTimeHydraulicStep, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_QUALSTEP)
    
    @property
    def getTimePatternStep(self):
        """ Retrieves the value of the pattern time step.
       
        Example:
          d.getTimePatternStep
       
        See also getTimePatternStart, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_PATTERNSTEP)
    
    @property
    def getTimePatternStart(self):
        """ Retrieves the value of pattern start time.
       
        Example:
          d.getTimePatternStart
       
        See also getTimePatternStep, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_PATTERNSTART)
    
    @property
    def getTimeReportingStep(self):
        """ Retrieves the value of the reporting time step.
       
        Example:
          d.getTimeReportingStep
       
        See also getTimeReportingPeriods, getTimeReportingStart.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_REPORTSTEP)
    
    @property
    def getTimeReportingStart(self):
        """ Retrieves the value of the reporting start time.
       
        Example:
          d.getTimeReportingStart
       
        See also getTimeReportingPeriods, getTimeReportingStep.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_REPORTSTART)
    
    @property
    def getTimeRuleControlStep(self):
        """ Retrieves the time step for evaluating rule-based controls.
       
        Example:
          d.getTimeRuleControlStep
       
        See also getTimeHydraulicStep.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_RULESTEP)
    
    @property
    def getTimeStatisticsType(self):
        """ Retrieves the type of time series post-processing.
       
        Types:
          1) NONE:    Reports the full time series for all quantities for all nodes and links (default)
          2) AVERAGE: Reports a set of time-averaged results
          3) MINIMUM: Reports only the minimum values
          3) MAXIMUM: Reports only the maximum values
          4) RANGE:   Reports the difference between the minimum and maximum values
       
        Example:
          d.getTimeStatisticsType
       
        See also getTimeStatisticsIndex, getTimeSimulationDuration.
        """
        self.TimeStatisticsIndex = self.d.ENgettimeparam(self.ToolkitConstants.EN_STATISTIC)
        return self.TYPESTATS[self.TimeStatisticsIndex]
    
    @property
    def getTimeStatisticsIndex(self):
        """ Retrieves the index of the type of time series post-processing.
       
        Type of time series post-processing:
          0 = 'NONE'
          1 = 'AVERAGE'
          2 = 'MINIMUM'
          3 = 'MAXIMUM'
          4 = 'RANGE'
       
        Example:
          d.getTimeStatisticsIndex
       
        See also getTimeStatisticsType, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_STATISTIC)
    
    @property
    def getTimeReportingPeriods(self):
        """ Retrieves the number of reporting periods saved to the binary.
       
        Example:
          d.getTimeReportingPeriods
       
        See also getTimeReportingStart, getTimeReportingStep.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_PERIODS)

    @property
    def getTimeStartTime(self):
        """ Retrieves the simulation starting time of day.
       
        Example:
          d.getTimeStartTime
       
        See also getTimeSimulationDuration, getTimePatternStart.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_STARTTIME)
    
    @property
    def getTimeHTime(self):
        """ Retrieves the elapsed time of current hydraulic solution.
       
        Example:
          d.getTimeHTime
       
        See also getTimeHydraulicStep, getComputedHydraulicTimeSeries.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_HTIME)
    
    @property
    def getTimeQTime(self):
        """Retrieves the elapsed time of current quality solution.
       
        Example:
          d.getTimeQTime
       
        See also getTimeQualityStep, getComputedQualityTimeSeries.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_QTIME)
    
    @property
    def getTimeHaltFlag(self):
        """ Retrieves the number of halt flag indicating if the simulation was halted.
       
        Example:
          d.getTimeHaltFlag
       
        See also getTimeStartTime, getTimeSimulationDuration.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_HALTFLAG)
        
    @property
    def getTimeNextEvent(self):
        """ Retrieves the shortest time until a tank becomes empty or full.
       
        Example:
          d.getTimeNextEvent
       
        See also getTimeNextEventTank.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_NEXTEVENT)
        
    @property
    def getTimeNextEventTank(self):
        """ Retrieves the index of tank with shortest time to become empty or full.
       
        Example:
          d.getTimeNextEventTank
       
        See also getTimeNextEvent.
        """
        return self.d.ENgettimeparam(self.ToolkitConstants.EN_NEXTEVENTTANK)
    
    def setLinkBulkReactionCoeff(self, value, *argv):
        """ Sets the value of bulk chemical reaction coefficient.
       
        Example 1:
          index_pipe = 1
          d.getLinkBulkReactionCoeff(index_pipe)              Retrieves the bulk chemical reaction coefficient of the 1st link
          coeff = 0
          d.setLinkBulkReactionCoeff(index_pipe, coeff)       Sets the bulk chemical reaction coefficient of the 1st link
          d.getLinkBulkReactionCoeff(index_pipe)
       
        Example 2:
          coeffs = d.getLinkBulkReactionCoeff()               Retrieves the bulk chemical reaction coefficients of all links
          coeffs_new = [0 for i in coeffs] 
          d.setLinkBulkReactionCoeff(coeffs_new)              Sets the bulk chemical reaction coefficient of all links
          d.getLinkBulkReactionCoeff()
       
        See also getLinkBulkReactionCoeff, setLinkRoughnessCoeff,
                 setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'KBULK', 'LINK', value, *argv)

    def setLinkDiameter(self, value, *argv):
        """ Sets the values of diameters.
       
        Example 1:
          d.getLinkDiameter()                           # Retrieves the diameters of all links
          index_pipe = 1
          diameter = 20
          d.setLinkDiameter(index_pipe, diameter)       # Sets the diameter of the 1st pipe
          d.getLinkDiameter(index_pipe)
       
        Example 2:
          index_pipes = [1, 2]
          diameters = [20, 25]
          d.setLinkDiameter(index_pipes, diameters)     # Sets the diameters of the first 2 pipes
          d.getLinkDiameter(index_pipes)
       
        Example 3:
          diameters = d.getLinkDiameter()
          diameters = [i * 1.5 for i in diameters]
          d.setLinkDiameter(diameters)                  # Sets the diameters of all the links
          d.getLinkDiameter()
       
        See also setLinkPipeData, setLinkLength,
                 setLinkBulkReactionCoeff, setLinkTypePipe.
        """
        self.set_eval('ENsetlinkvalue', 'DIAMETER', 'LINK', value, *argv)
      
    def setLinkInitialSetting(self, value, *argv):
        """ Sets the values of initial settings, roughness for pipes or initial speed for pumps or initial setting for valves.
       
        Example 1:
          index_pipe = 1
          d.getLinkInitialSetting(index_pipe)                 # Retrieves the initial setting of the 1st link
          setting = 80
          d.setLinkInitialSetting(index_pipe, setting)        # Sets the initial setting of the 1st link
          d.getLinkInitialSetting(index_pipe)
       
        Example 2:
          settings = d.getLinkInitialSetting()                # Retrieves the initial setting of all links
          settings_new = [i + 40 for i in settings]
          d.setLinkInitialSetting(settings_new)               # Sets the initial setting of all links
          d.getLinkInitialSetting()
       
        See also getLinkInitialSetting, setLinkInitialStatus, setLinkRoughnessCoeff,
                 setLinkPipeData, addLink, deleteLink. 
        """
        self.set_eval('ENsetlinkvalue', 'INITSETTING', 'LINK', value, *argv)
 
    def setLinkInitialStatus(self, value, *argv):
        """ Sets the values of initial status.
       
        Note: Cannot set status for a check valve
       
        Example 1:
          index_pipe = 1
          d.getLinkInitialStatus(index_pipe)                # Retrieves the initial status of the 1st link
          status = 0
          d.setLinkInitialStatus(index_pipe, status)        # Sets the initial status of the 1st link
          d.getLinkInitialStatus(index_pipe)
       
        Example 2:
          statuses = d.getLinkInitialStatus                 # Retrieves the initial status of all links
          statuses_new = zeros(1, length(statuses))
          d.setLinkInitialStatus(statuses_new)              # Sets the initial status of all links
          d.getLinkInitialStatus
       
        See also getLinkInitialStatus, setLinkInitialSetting, setLinkDiameter,
                 setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'INITSTATUS', 'LINK', value, *argv)
         
    def setLinkLength(self, value, *argv):
        """ Sets the values of lengths.
         
          Example 1:
            index_pipe = 1
            d.getLinkLength(index_pipe)                   # Retrieves the length of the 1st link
            length_pipe = 100
            d.setLinkLength(index_pipe, length_pipe)      # Sets the length of the 1st link
            d.getLinkLength(index_pipe)
         
          Example 2:
            lengths = d.getLinkLength()                   # Retrieves the lengths of all the links
            lengths_new = [i * 1.5 for i in lengths]
            d.setLinkLength(lengths_new)                  # Sets the new lengths of all links
            d.getLinkLength()
         
          See also getLinkLength, setLinkDiameter, setLinkMinorLossCoeff, 
                   setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'LENGTH','LINK', value, *argv)
     
    def setLinkMinorLossCoeff(self, value, *argv):
        """ Sets the values of minor loss coefficient.
       
        Example 1:
          index_pipe = 1
          d.getLinkMinorLossCoeff(index_pipe)               # Retrieves the minor loss coefficient of the 1st link
          coeff = 105
          d.setLinkMinorLossCoeff(index_pipe, coeff)        # Sets the minor loss coefficient of the 1st link
          d.getLinkMinorLossCoeff(index_pipe)
       
        Example 2:
          coeffs = d.getLinkMinorLossCoeff                  # Retrieves the minor loss coefficients of all the links
          coeffs_new = coeffs + 0.2
          d.setLinkMinorLossCoeff(coeffs_new)               # Sets the minor loss coefficient of all links
          d.getLinkMinorLossCoeff()
       
        See also getLinkMinorLossCoeff, setLinkDiameter, setLinkRoughnessCoeff,
                 setLinkPipeData, addLink, deleteLink.    
        """
        self.set_eval('ENsetlinkvalue', 'MINORLOSS','LINK', value, *argv)
   
    def setLinkRoughnessCoeff(self, value, *argv):
        """ Sets the values of roughness coefficient.
       
        Example 1:
          index_pipe = 1
          d.getLinkRoughnessCoeff(index_pipe)               # Retrieves the roughness coefficient of the 1st link
          coeff = 105 
          d.setLinkRoughnessCoeff(index_pipe, coeff)        # Sets the roughness coefficient of the 1st link
          d.getLinkRoughnessCoeff(index_pipe)
       
        Example 2:
          coeffs = d.getLinkRoughnessCoeff                  # Retrieves the roughness coefficients of all the links
          coeffs_new = coeffs + 10 
          d.setLinkRoughnessCoeff(coeffs_new)               # Sets the roughness coefficient of all links
          d.getLinkRoughnessCoeff
       
        See also getLinkRoughnessCoeff, setLinkDiameter, setLinkMinorLossCoeff,
                 setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'ROUGHNESS','LINK', value, *argv)
        
    def setLinkSettings(self, value, *argv):
        """Sets the values of current settings, roughness for pipes or initial speed for pumps or initial setting for valves.
       
        Example 1:
          index_pipe = 1
          d.getLinkSettings(index_pipe)                 # Retrieves the current setting of the 1st link
          setting = 80
          d.setLinkSettings(index_pipe, setting)        # Sets the current setting of the 1st link
          d.getLinkSettings(index_pipe)
       
        Example 2:
          settings = d.getLinkSettings()                # Retrieves the current setting of all links
          settings_new = [i + 40 for i in settings]
          d.setLinkSettings(settings_new)               # Sets the current setting of all links
          d.getLinkSettings()
       
        See also getLinkSettings, setLinkStatus, setLinkRoughnessCoeff,
                 setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'SETTING','LINK', value, *argv)

    def setLinkStatus(self, value, *argv):
        """ Sets the values of current status for links.
           
        Note: Cannot set status for a check valve
           
        Example 1:
          index_pipe = 1
          d.getLinkStatus(index_pipe)                  # Retrieves the current status of the 1st link
          status = 1
          d.setLinkStatus(index_pipe, status)          # Sets the current status of the 1st link
          d.getLinkStatus(index_pipe)
           
        Example 2:
          statuses = d.getLinkStatus()                 # Retrieves the current status of all links
          statuses_new = [0 for i in statuses]
          d.setLinkStatus(statuses_new)                # Sets the current status of all links
          d.getLinkStatus()
           
        See also getLinkStatus, setLinkInitialStatus, setLinkDiameter,
                 setLinkPipeData, addLink, deleteLink.
        """
        self.set_eval('ENsetlinkvalue', 'STATUS','LINK', value, *argv)

    def setLinkWallReactionCoeff(self, value, *argv):
        """ Sets the value of wall chemical reaction coefficient.
       
        Example 1:
          index_pipe = 1
          d.getLinkWallReactionCoeff(index_pipe)               # Retrieves the wall chemical reaction coefficient of the 1st link
          coeff = 0
          d.setLinkWallReactionCoeff(index_pipe, coeff)        # Sets the wall chemical reaction coefficient of the 1st link
          d.getLinkWallReactionCoeff(index_pipe)
       
        Example 2:
          coeffs = d.getLinkWallReactionCoeff()                # Retrieves the wall chemical reaction coefficients of all links
          coeffs_new = [0 for i in coeffs]
          d.setLinkWallReactionCoeff(coeffs_new)               # Sets the wall chemical reaction coefficient of all links
          d.getLinkWallReactionCoeff()
       
        See also getLinkWallReactionCoeff, setLinkBulkReactionCoeff,
                 setLinkPipeData, addLink, deleteLink.    
        """
        self.set_eval('ENsetlinkvalue', 'KWALL','LINK', value, *argv)     
 
    def setNodeElevations(self, value, *argv):
        """ Sets the values of elevation for nodes.
         
        Example 1:
          index_node = 1 
          d.getNodeElevations(index_node)            # Retrieves the elevation of the 1st node
          elev = 500 
          d.setNodeElevations(index_node, elev)      # Sets the elevation of the 1st node
          d.getNodeElevations(index_node)
       
        Example 2:
          elevs = d.getNodeElevations()               # Retrieves the elevations of all the nodes
          elevs_new = [i + 100 for i in elevs]
          d.setNodeElevations(elevs_new)              # Sets the elevations of all nodes
          d.getNodeElevations()
       
        See also getNodeElevations, setNodeCoordinates, setNodeBaseDemands,
                 setNodeJunctionData, addNodeJunction, deleteNode.
        """
        self.set_eval('ENsetnodevalue', 'ELEVATION','NODE', value, *argv)
        
    def setNodeTankBulkReactionCoeff(self, value, *argv):
        """ Sets the tank bulk reaction coefficient.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankBulkReactionCoeff()                          # Retrieves the  bulk reaction coefficient of all tanks
          d.setNodeTankBulkReactionCoeff(-0.5)                      # Sets the bulk reaction coefficient = -0.5 to every tank
          d.getNodeTankBulkReactionCoeff()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankBulkReactionCoeff([0, -0.5])                 # Sets the bulk reaction coefficient = 0 and -0.5 to the 2 tanks
          d.getNodeTankBulkReactionCoeff()
       
        Example 3:
          d.setNodeTankBulkReactionCoeff(1, -0.8)                   # Sets the bulk reaction coefficient = -0.5 to the 1st tank
          d.getNodeTankBulkReactionCoeff()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankBulkReactionCoeff(tankIndex, 0)              # Sets the bulk reaction coefficient = 0 to the tanks with index 128 and 129
          d.getNodeTankBulkReactionCoeff()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankBulkReactionCoeff(tankIndex, [-0.5, 0])      # Sets the bulk reaction coefficient = -0.5 and 0 to the tanks with index 128 and 129 respectively
          d.getNodeTankBulkReactionCoeff()
       
        See also getNodeTankBulkReactionCoeff, setNodeTankInitialLevel, setNodeTankMixingModelType,
                 setNodeTankCanOverFlow, setNodeTankDiameter, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'TANK_KBULK','TANK', value, *argv)
          
    def setNodeTankDiameter(self, value, *argv):
        """ Sets the diameter value for tanks.
       
        The examples are based on d=epanet('BWSN_Network_1.inp')
       
        Example 1:
          d.getNodeTankDiameter()                         #  Retrieves the diameter of all tanks
          d.setNodeTankDiameter(120)                      #  Sets the diameter = 120 to every tank
          d.getNodeTankDiameter()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankDiameter([110, 130])               # Sets the diameter = 110 and 130 to the 2 tanks
          d.getNodeTankDiameter()
       
        Example 3:
          d.setNodeTankDiameter(1, 120)                   # Sets the diameter = 120 to the 1st tank
          d.getNodeTankDiameter()
-       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankDiameter(tankIndex, 150)           # Sets the diameter = 150 to the tanks with index 128 and 129
          d.getNodeTankDiameter()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankDiameter(tankIndex, [100, 120])    # Sets the diameter = 100 and 120 to the tanks with index 128 and 129 respectively
          d.getNodeTankDiameter()
       
        See also getNodeTankDiameter, setNodeTankInitialLevel, setNodeTankMinimumWaterLevel,
                 setNodeTankBulkReactionCoeff, setNodeTankCanOverFlow, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'TANKDIAM','TANK', value, *argv)
                
    def setNodeTankInitialLevel(self, value, *argv):
        """ Sets the values of initial level for tanks.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankInitialLevel()                       # Retrieves the initial level of all tanks
          d.setNodeTankInitialLevel(10)                     # Sets the initial level = 10 to every tank
          d.getNodeTankInitialLevel()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankInitialLevel([10, 15])               # Sets the initial level = 10 and 15 to the 2 tanks
          d.getNodeTankInitialLevel()
       
        Example 3:
          d.setNodeTankInitialLevel(1, 10)                  # Sets the initial level = 10 to the 1st tank
          d.getNodeTankInitialLevel()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankInitialLevel(tankIndex, 10)          # Sets the initial level = 10 to the tanks with index 128 and 129
          d.getNodeTankInitialLevel()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankInitialLevel(tankIndex, [10, 15])    # Sets the initial level = 10 and 15 to the tanks with index 128 and 129 respectively
          d.getNodeTankInitialLevel()
       
        See also getNodeTankInitialLevel, setNodeTankMinimumWaterLevel, setNodeTankMaximumWaterLevel,
                  setNodeTankMinimumWaterVolume, setNodeTankMixingFraction, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'TANKLEVEL','TANK', value, *argv)
        
    def setNodeTankMaximumWaterLevel(self, value, *argv):
        """ Sets the maximum water level value for tanks.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankMaximumWaterLevel()                       # Retrieves the maximum water level of all tanks
          d.setNodeTankMaximumWaterLevel(35)                     # Sets the maximum water level = 35 to every tank
          d.getNodeTankMaximumWaterLevel()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankMaximumWaterLevel([30, 40])               # Sets the maximum water level = 30 and 40 to the 2 tanks
          d.getNodeTankMaximumWaterLevel()
       
        Example 3:
          d.setNodeTankMaximumWaterLevel(1, 35)                  # Sets the maximum water level = 35 to the 1st tank
          d.getNodeTankMaximumWaterLevel()
       
        Example 4:
          tankIndex = d.getNodeTankIndex;
          d.setNodeTankMaximumWaterLevel(tankIndex, 30)          # Sets the maximum water level = 30 to the tanks with index 128 and 129
          d.getNodeTankMaximumWaterLevel()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMaximumWaterLevel(tankIndex, [35, 45])    # Sets the maximum water level = 35 and 45 to the tanks with index 128 and 129 respectively
          d.getNodeTankMaximumWaterLevel()
       
        See also getNodeTankMaximumWaterLevel, setNodeTankInitialLevel, setNodeTankMinimumWaterLevel,
                 setNodeTankMinimumWaterVolume, setNodeTankMixingFraction, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'MAXLEVEL','TANK', value, *argv)
        
    def setNodeTankMinimumWaterLevel(self, value, *argv):
        """ Sets the minimum water level value for tanks.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankMinimumWaterLevel()                         # Retrieves the minimum water level of all tanks
          d.setNodeTankMinimumWaterLevel(5)                        # Sets the minimum water level = 5 to every tank
          d.getNodeTankMinimumWaterLevel()
       
        Example 2:
             The input array must have a length equal to the number of tanks
          d.setNodeTankMinimumWaterLevel([10, 15])                 # Sets the minimum water level = 10 and 15 to the 2 tanks
          d.getNodeTankMinimumWaterLevel()
       
        Example 3:
          d.setNodeTankMinimumWaterLevel(1, 5)                     # Sets the minimum water level = 5 to the 1st tank
          d.getNodeTankMinimumWaterLevel()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMinimumWaterLevel(tankIndex, 10)            # Sets the minimum water level = 10 to the tanks with index 128 and 129
          d.getNodeTankMinimumWaterLevel()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMinimumWaterLevel(tankIndex, [5, 15])       # Sets the minimum water level = 5 and 15 to the tanks with index 128 and 129 respectively
          d.getNodeTankMinimumWaterLevel()
       
        See also getNodeTankMinimumWaterLevel, setNodeTankInitialLevel, setNodeTankMaximumWaterLevel,
                 setNodeTankMinimumWaterVolume, setNodeTankMixingFraction, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'MINLEVEL','TANK', value, *argv)
    
    def setNodeTankMinimumWaterVolume(self, value, *argv):
        """ Sets the minimum water volume value for tanks.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankMinimumWaterVolume()                           # Retrieves the minimum water volume of all tanks
          d.setNodeTankMinimumWaterVolume(1000)                       # Sets the minimum water volume = 1000 to every tank
          d.getNodeTankMinimumWaterVolume()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankMinimumWaterVolume([1500, 2000])               # Sets the minimum water volume = 1500 and 2000 to the 2 tanks
          d.getNodeTankMinimumWaterVolume()
       
        Example 3:
          d.setNodeTankMinimumWaterVolume(1, 1000)                    # Sets the minimum water volume = 1000 to the 1st tank
          d.getNodeTankMinimumWaterVolume()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMinimumWaterVolume(tankIndex, 1500)            # Sets the minimum water volume = 1500 to the tanks with index 128 and 129
          d.getNodeTankMinimumWaterVolume()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMinimumWaterVolume(tankIndex, [1000, 2000])     # Sets the minimum water volume = 1000 and 2000 to the tanks with index 128 and 129 respectively
          d.getNodeTankMinimumWaterVolume()
       
        See also getNodeTankMinimumWaterVolume, setNodeTankInitialLevel, setNodeTankMinimumWaterLevel,
                 setNodeTankMaximumWaterLevel, setNodeTankMixingFraction, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'MINVOLUME','TANK', value, *argv)
        
    def setNodeTankMixingFraction(self, value, *argv):
        """ Sets the tank mixing fraction of total volume occupied by the inlet/outlet zone in a 2-compartment tank.
       
        The examples are based on d=epanet('BWSN_Network_1.inp');
       
        Example 1:
          d.getNodeTankMixingFraction()                     # Retrieves the mixing fraction of all tanks
          d.setNodeTankMixingFraction(0)                    # Sets the mixing fraction = 0 to every tank
          d.getNodeTankMixingFraction()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankMixingFraction([1, 0])               # Sets the mixing fraction = 1 and 0 to the 2 tanks
          d.getNodeTankMixingFraction()
       
        Example 3:
          d.setNodeTankMixingFraction(1, 0)                 # Sets the mixing fraction = 0 to the 1st tank
          d.getNodeTankMixingFraction()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMixingFraction(tankIndex, 1)         # Sets the mixing fraction = 1 to the tanks with index 128 and 129
          d.getNodeTankMixingFraction()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMixingFraction(tankIndex, [1, 0])     # Sets the mixing fraction = 1 and 0 to the tanks with index 128 and 129 respectively
          d.getNodeTankMixingFraction()
       
        See also getNodeTankMixingFraction, setNodeTankMixingModelType, setNodeTankMinimumWaterLevel,
                 setNodeTankMinimumWaterVolume, setNodeTankDiameter, setNodeTankData.
        """
        self.set_eval_link_node('ENsetnodevalue', 'MIXFRACTION','TANK', value, *argv)
        
    def setNodeTankMixingModelType(self, value, *argv):
        """ Sets the mixing model type value for tanks.
       
        The examples are based on d = epanet('BWSN_Network_1.inp')
       
        Example 1:
          d.getNodeTankMixingModelType()                                 # Retrieves the  mixing model type of all tanks
          d.setNodeTankMixingModelType('MIX2')                           # Sets the  mixing model type = 'MIX2' to every tank
          d.getNodeTankMixingModelType()
       
        Example 2:
            The input array must have a length equal to the number of tanks
          d.setNodeTankMixingModelType(['MIX1', 'LIFO'])                 # Sets the  mixing model type = 'MIX1' and 'LIFO' to the 2 tanks
          d.getNodeTankMixingModelType()
       
        Example 3:
          d.setNodeTankMixingModelType(1, 'FIFO')                        # Sets the  mixing model type = 'FIFO' to the 1st tank
          d.getNodeTankMixingModelType()
       
        Example 4:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMixingModelType(tankIndex, 'MIX1')                # Sets the  mixing model type = 'MIX1' to the tanks with index 128 and 129
          d.getNodeTankMixingModelType()
       
        Example 5:
          tankIndex = d.getNodeTankIndex()
          d.setNodeTankMixingModelType(tankIndex, ['MIX2', 'LIFO'])      # Sets the  mixing model type = 'MIX2' and 'LIFO' to the tanks with index 128 and 129 respectively
          d.getNodeTankMixingModelType()
       
        See also getNodeTankMixingModelType, setNodeTankBulkReactionCoeff, setNodeTankMixingFraction,
                 setNodeTankMinimumWaterVolume, setNodeTankMinimumWaterLevel, setNodeTankData.
        """
        if len(argv) == 0:
            _type = value
        elif len(argv) == 1:
            _type = argv[0]
        if type(_type) is list:
            code = []
            for i in _type:
                code.append(self.TYPEMIXMODEL.index(i))
        else:
            code = self.TYPEMIXMODEL.index(_type)
        if len(argv) == 0:
            value = code
        elif len(argv) == 1:
            argv = code
            self.set_eval_link_node('ENsetnodevalue', 'MIXMODEL','TANK', value, argv)
            return
        self.set_eval_link_node('ENsetnodevalue', 'MIXMODEL','TANK', value, *argv)
        
    def setTimeHydraulicStep(self, value):
        """ Sets the hydraulic time step.
       
        Example:
          Hstep = 1800
          d.setTimeHydraulicStep(Hstep)
          d.getTimeHydraulicStep
       
        See also getTimeHydraulicStep, setTimeQualityStep, setTimePatternStep.
        """
        self.d.ENsettimeparam(self.ToolkitConstants.EN_HYDSTEP, value)

    def setTimeQualityStep(self, value):
        """ Sets the quality time step.
       
        Example:
          Qstep = 1800
          d.setTimeQualityStep(Qstep)
          d.getTimeQualityStep
       
        See also getTimeQualityStep, setTimeHydraulicStep, setTimePatternStep.
        """
        self.d.ENsettimeparam(self.ToolkitConstants.EN_QUALSTEP, value)
        
    def setTimeReportingStep(self, value):
        """ Sets the reporting time step.
       
        Example:
          reportingStep = 3600
          d.setTimeReportingStep(reportingStep)
          d.getTimeReportingStep
       
        See also getTimeReportingStep, setTimeReportingStart, setTimeRuleControlStep.
        """
        self.d.ENsettimeparam(self.ToolkitConstants.EN_REPORTSTEP, value)
        
    def setTimeSimulationDuration(self, value):
        """ Sets the simulation duration (in seconds).
       
        Example:
          simulationDuration = 172800    # 172800 seconds = 2days
          d.setTimeSimulationDuration(simulationDuration)
          d.getTimeSimulationDuration()
       
        See also getTimeSimulationDuration, getTimeStartTime, getTimeHaltFlag.
        """
        self.d.ENsettimeparam(self.ToolkitConstants.EN_DURATION, value)
              
    def getPatternNameID(self, *argv):
        """ Retrieves the ID label of all or some time patterns indices.
       
        Example 1:
          d.getPatternNameID()          # Retrieves the IDs of all the patterns
       
        Example 2:
          d.getPatternNameID(1)         # Retrieves the ID of the 1st pattern
       
        Example 3:
          d.getPatternNameID([1,2])     # Retrieves the IDs of the first 2 patterns
       
        See also setPatternNameID, getPattern.
        """
        try:
            values = [] 
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetpatternid(i))
                else:
                    values = self.d.ENgetpatternid(index) 
            else:
                for i in range(self.getPatternCount()):
                    values.append(self.d.ENgetpatternid(i+1))
        except Exception as Errcode:
            return self.getError(Errcode)
        return values
    
    @property 
    def closeHydraulicAnalysis(self):
        """ Closes the hydraulic analysis system, freeing all allocated memory.
       
        Example:
          d.closeHydraulicAnalysis
       
        For more, you can type `help getNodePressure` and check examples 3 & 4.
       
        See also openHydraulicAnalysis, saveHydraulicFile, closeQualityAnalysis.
        """
        self.d.ENcloseH()
        
    def initializeHydraulicAnalysis(self, *argv):
        """ Initializes storage tank levels, link status and settings, and the simulation clock time prior to running a hydraulic analysis.
         
        Codes:
          1) NOSAVE        = 0,    Don't save hydraulics  don't re-initialize flows
          2) SAVE          = 1,    Save hydraulics to file, don't re-initialize flows
          3) INITFLOW      = 10,   Don't save hydraulics  re-initialize flows
          4) SAVE_AND_INIT = 11    Save hydraulics  re-initialize flows
         
        Example 1:
          d.initializeHydraulicAnalysis()     # Uses the default code i.e. SAVE = 1
         
        Example 2:
          code = 0                         i.e. Don't save
          d.initializeHydraulicAnalysis(code)
         
        For more, you can type `help getNodePressure` and check examples 3 & 4.
         
        See also saveHydraulicFile, initializeQualityAnalysis.
        """
        try:
            code = self.ToolkitConstants.EN_SAVE
            if len(argv)>0:
                code = argv[0]
            value = self.d.ENinitH(code)
        except Exception as Errcode:
            return self.getError(Errcode)
        return value
        
    @property
    def openHydraulicAnalysis(self):
        """ Opens the hydraulics analysis system.
           
        Example:
          d.openHydraulicAnalysis
           
        For more, you can type `help getNodePressure` and check examples 3 & 4.
           
        See also openQualityAnalysis, initializeHydraulicAnalysis.
        """
        try:
            self.d.ENopenH()
        except Exception as Errcode:
            return self.getError(Errcode) 
    
    @property
    def runHydraulicAnalysis(self):
        """ Runs a single period hydraulic analysis, retrieving the current simulation clock time t.
         
        Example:
          tstep = d.runHydraulicAnalysis()
         
        For more, you can type `help getNodePressure` and check examples 3 & 4.
         
        See also runQualityAnalysis, initializeHydraulicAnalysis.
        """
        try:
            value = self.d.ENrunH()
        except Exception as Errcode:
            return self.getError(Errcode) 
        return value
    
    @property
    def nextHydraulicAnalysisStep(self):
        """ Determines the length of time until the next hydraulic event occurs in an extended period simulation.
       
        Example:
          d.nextHydraulicAnalysisStep
       
        For more, you can type `help getNodePressure` and check examples 3 & 4.
       
        See also nextQualityAnalysisStep, runHydraulicAnalysis.
        """
        return self.d.ENnextH()
     
    ######### Lower Functions ############        
    
    def get_link_info(self, iCode, *argv):
        try:
            values = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    for i in index:
                        values.append(self.d.ENgetlinkvalue(i, iCode)) 
                else:
                    values = self.d.ENgetlinkvalue(index, iCode)
            else:
                for i in range(self.getLinkCount()):
                    values.append(self.d.ENgetlinkvalue(i+1, iCode))
        except Exception as Errcode:
            return self.getError(Errcode) 
        return values
    
    def get_node_info(self, iCode, *argv):
        try:
             value = []
             if len(argv) > 0:
                 index = argv[0]
                 if type(index) is list:
                     for i in index:
                         value.append(self.d.ENgetnodevalue(i, iCode)) 
                 else:
                     value = self.d.ENgetnodevalue(index, iCode) 
             else:
                 for i in range(self.getNodeCount()):
                     value.append(self.d.ENgetnodevalue(i+1, iCode))
        except Exception as Errcode:
            return self.getError(Errcode)
        return value
    
    def get_node_tank_mixining_model(self, *argv):  
        self.NodeTankMixingModelCode = self.get_tank_node_info(self.ToolkitConstants.EN_MIXMODEL, *argv)
        if type(self.NodeTankMixingModelCode) is list:
            self.NodeTankMixingModelType = [self.TYPEMIXMODEL[int(i)] for i in self.NodeTankMixingModelCode]
        else:
            self.NodeTankMixingModelType = self.TYPEMIXMODEL[int(self.NodeTankMixingModelCode)] 
        return [self.NodeTankMixingModelCode, self.NodeTankMixingModelType]

    def get_pump_link_info(self, iCode, *argv):
        # check if the *argv input is the real index of the pumps
        indices = self.getLinkPumpIndex()
        try:
            values = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    if index != indices:
                        index = self.getLinkPumpIndex(index)
                    for i in index:
                        values.append(self.d.ENgetlinkvalue(i, iCode)) 
                else:
                    if [index] != indices:
                        pIndex = self.getLinkPumpIndex(index)
                    values = self.d.ENgetlinkvalue(pIndex, iCode)
            else:
                for i in indices:
                    values.append(self.d.ENgetlinkvalue(i, iCode))
        except Exception as Errcode:
            return self.getError(Errcode) 
        return values 
    
    def get_tank_node_info(self, iCode, *argv):
        indices = self.getNodeTankIndex()
        try:
            values = []
            if len(argv) > 0:
                index = argv[0]
                if type(index) is list:
                    if index != indices:
                        index = self.getNodeTankIndex(index)
                    for i in index:
                        values.append(self.d.ENgetnodevalue(i, iCode)) 
                else:
                    if [index] != indices:
                        index = self.getNodeTankIndex(index)
                    values = self.d.ENgetnodevalue(index, iCode)
            else:
                for i in indices:
                    values.append(self.d.ENgetnodevalue(i, iCode))
        except Exception as Errcode:
            return self.getError(Errcode) 
        return values 

    def set_eval(self, func, iCodeStr, Type, value, *argv):
        try:
            if len(argv) == 1:
                index = value
                value = argv[0]
                if type(index) is list:
                    j = 0
                    for i in index:
                        strFunc = 'self.d.' + func + '(' + str(i)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value[j]) +')'
                        eval(strFunc)
                        j+=1    
                else:
                    strFunc = 'self.d.' + func + '(' + str(index)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value) +')'
                    eval(strFunc)
            else:
                count = 0
                if (Type == 'LINK'):
                    count = self.getLinkCount()
                elif (Type == 'NODE'):
                    count = self.getNodeCount()
                for i in range(count):
                    strFunc = 'self.d.' + func + '(' + str(i+1)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value[i]) +')'
                    eval(strFunc)
        except Exception as Errcode:
            return self.getError(Errcode) 
        
    def set_eval_link_node(self, func, iCodeStr, Type, value, *argv):
        try:
            if len(argv) == 1:
                index = value
                value = argv[0]
                if type(index) is list:
                    j = 0
                    if type(value) is list:
                        for i in index:
                            strFunc = 'self.d.' + func + '(' + str(i)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value[j]) +')'
                            eval(strFunc)
                            j+=1 
                    else:
                        for i in index:
                            strFunc = 'self.d.' + func + '(' + str(i)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value) +')'
                            eval(strFunc)
                            j+=1                         
                else:
                    try:
                        tIndex = self.getNodeTankIndex(index)
                    except:
                        tIndex = index
                    strFunc = 'self.d.' + func + '(' + str(tIndex)+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value) +')'
                    eval(strFunc)
            else:
                count = 0
                if (Type == 'TANK'):
                    count = self.getNodeTankCount()
                    indices = self.getNodeTankIndex()
                elif (Type == 'PUMP'):
                    count = self.getLinkPumpCount()
                    indices = self.getLinkPumpIndex()
                if type(value) is list:
                    for i in range(count):
                        strFunc = 'self.d.' + func + '(' + str(indices[i])+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value[i]) +')'
                        eval(strFunc)
                else:
                    for i in range(count):
                        strFunc = 'self.d.' + func + '(' + str(indices[i])+ ',' + 'self.ToolkitConstants.EN_' + iCodeStr + ',' + str(value) +')'
                        eval(strFunc)
                    
        except Exception as Errcode:
            return self.getError(Errcode) 
        
    def getlinkIndices(self, *argv):
        if len(argv)>0:
            if  type(argv[0]) is list:
                return argv[0]
            else:
                return [argv[0]]
        else:
            return self.getLinkIndex()
       
    def getNodeIndices(self, *argv):
        if len(argv)>0:
            if  type(argv[0]) is list:
                return argv[0]
            else:
                return [argv[0]]
        else:
            return self.getNodeIndex()
        
    def plot(self, *argv):
        """ Plot Network, show all components, plot pressure
       
        Example 1:
          d = epanet('Net1.inp')  
          d.plot()                                   # Plot Net1.inp network 
          
        Example 2:
          d = epanet('Net1.inp')                     # Run hydralic analysis and plot the pressures at 10hrs 
          d.openHydraulicAnalysis
          d.initializeHydraulicAnalysis() 
          tstep, P = 1, []
          while (tstep>0):
             t = d.runHydraulicAnalysis 
             P.append(d.getNodePressure())
             tstep=d.nextHydraulicAnalysisStep 
          d.closeHydraulicAnalysis
          hr = 10
          d.plot('Pressure', P[hr])                   
        """
        # get info from EN functions
        nodenameid = self.getNodeNameID()
        linknameid = self.getLinkNameID()  
        if len(nodenameid) == 0 or len(linknameid) == 0:
            raise Epanet_Except('Not enough network nodes/links.') 
        nodesconnlinks    = self.getNodesConnectingLinksID() 
        nodeconlinkIndex  = self.getNodesConnectingLinksIndex() 
        pumpindex         = self.getLinkPumpIndex() 
        valveindex        = self.getLinkValveIndex() 
        juncind           = self.getNodeJunctionIndex()
        resindex          = self.getNodeReservoirIndex() 
        tankindex         = self.getNodeTankIndex()
        linkcount         = self.getLinkCount() 
        nodecount         = self.getNodeCount()
        linkindex         = self.getLinkIndex() 
        nodeindex         = self.getNodeIndex()
        nodecoords        = self.getNodeCoordinates()
        
        # Create figure
        plt.rcParams['figure.dpi'] = 300
        # fig, pl = plt.subplot()
        if len(argv) == 0:
            title = os.path.basename(self.inp_file)
            plt.title(title[0:len(title)-4])
        plt.axis('off')
        
        # # Set Limits
        # xMax = max(nodecoords['x'].values()) 
        # xMin = min(nodecoords['x'].values()) 
        # yMax = max(nodecoords['y'].values())
        # yMin = min(nodecoords['y'].values())
        # avgX = abs(xMax-xMin)/2
        # avgY = abs(yMax-yMin)/2
        # margin_param = 0.2
        # xMargMin = xMin - margin_param * avgX
        # xMargMax = xMax + margin_param * avgX
        # yMargMin = yMin - margin_param * avgY
        # yMargMax = yMax + margin_param * avgY
        # plt.xlim([xMargMin, xMargMax])
        # plt.ylim([yMargMin, yMargMax]) 
    
        # Plot Links 
        x, y = [0, 0], [0,0]
        for i in range(linkcount):
            fromNode = nodeconlinkIndex[i][0]
            toNode   = nodeconlinkIndex[i][1]
            x[0] = nodecoords['x'][fromNode]
            y[0] = nodecoords['y'][fromNode]
            x[1] = nodecoords['x'][toNode]
            y[1] = nodecoords['y'][toNode]
            if nodecoords['x_vert'][i+1] == []:
                plt.plot(x, y, 'k-', linewidth = 0.5, zorder=1)  # label ='Links')
            else:
                xV_old = x[0]
                yV_old = y[0]
                for j in range(len(nodecoords['x_vert'][i+1])):
                      xV =  nodecoords['x_vert'][i+1][j]         
                      yV =  nodecoords['y_vert'][i+1][j]   
                      plt.plot([xV_old, xV], [yV_old, yV], 'k-', linewidth = 0.5)
                      xV_old = xV
                      yV_old = yV
                plt.plot([xV, x[1]], [yV, y[1]], 'k-', linewidth = 0.5, zorder=1)
        
        if len(argv) == 0:        
            # Plot Pumps
            for i in pumpindex:
                fromNode = nodeconlinkIndex[i-1][0]
                toNode   = nodeconlinkIndex[i-1][1]
                x[0] = nodecoords['x'][fromNode]
                y[0] = nodecoords['y'][fromNode]
                x[1] = nodecoords['x'][toNode]
                y[1] = nodecoords['y'][toNode]
                xx = (x[0] + x[1]) / 2
                yy = (y[0] + y[1]) / 2
                if i != pumpindex[-1]:
                    plt.plot(xx, yy, 'vm', markersize = 5)
                else:
                    plt.plot(xx, yy, 'vm', markersize = 5, label = 'Pumps')
                    
            # Plot Valves
            for i in valveindex:
                fromNode = nodeconlinkIndex[i-1][0]
                toNode   = nodeconlinkIndex[i-1][1]
                x[0] = nodecoords['x'][fromNode]
                y[0] = nodecoords['y'][fromNode]
                x[1] = nodecoords['x'][toNode]
                y[1] = nodecoords['y'][toNode]
                xx = (x[0] + x[1]) / 2    # IS IT PLACED IN THE MIDDLE?????????????
                yy = (y[0] + y[1]) / 2
                if i != valveindex[-1]:
                    plt.plot(xx, yy, 'k*', markersize = 5)
                else:
                    plt.plot(xx, yy, 'k*', markersize = 5, label = 'Valves')
                    
            # Plot Junctions
            for i in juncind:
                x = nodecoords['x'][i]
                y = nodecoords['y'][i]
                if i != juncind[-1]:
                    plt.plot(x, y, 'bo', markersize = 2)
                else:
                    plt.plot(x, y, 'bo', markersize = 2, label = 'Junctions')
                    
            # Plot Reservoirs
            for i in resindex:
                x = nodecoords['x'][i]
                y = nodecoords['y'][i]
                if i != resindex[-1]:
                    plt.plot(x, y, 'gs', markersize = 4) 
                else:
                    plt.plot(x, y, 'gs', markersize = 4, label = 'Reservoirs') 
                    
            # Plot Tanks
            for i in tankindex:
                x = nodecoords['x'][i]
                y = nodecoords['y'][i]
                if i != tankindex[-1]:
                    plt.plot(x, y, '*c', markersize = 8)  
                else:
                    plt.plot(x, y, '*c', markersize = 8, label = 'Tanks')  
            plt.legend(loc = 0, fontsize = 7)
        else:    
            # Plot Pressure
            if argv[0] == 'Pressure':
                if len(argv) >1 and type(argv[1]) is not str:
                    Pressure = argv[1]
                else:
                    Pressure = self.getNodePressure()
                # plot pressure nodes        
                x = list(nodecoords['x'].values())
                y = list(nodecoords['y'].values())
                plt.scatter(x, y, c = Pressure, cmap="jet", s=5, zorder=2)
    
                # Plot details
                bar = plt.colorbar(orientation = 'horizontal',fraction=0.15, shrink = 0.9, pad=0.01, aspect = 50)
                bar.set_label(label='Pressure (psi)', size =10) 
                title = os.path.basename(self.inp_file)
                plt.title(title[0:len(title)-4]+': Pressure')  
                
            # Plot Elevation
            if argv[0] == 'Elevation':
                if len(argv) >1 and type(argv[1]) is not str:
                    Elevations = argv[1]
                else:
                    Elevations = self.getNodeElevations()
                # plot pressure nodes        
                x = list(nodecoords['x'].values())
                y = list(nodecoords['y'].values())
                plt.scatter(x, y, c = Elevations, cmap="jet", s=5, zorder=2)
                # Plot details
                bar = plt.colorbar(orientation = 'horizontal',fraction=0.15, shrink = 0.9, pad=0.01, aspect = 50)
                bar.set_label(label = 'Elevation (feet)', size =10) 
                title = os.path.basename(self.inp_file)
                plt.title(title[0:len(title)-4]+': Elevation')  
        plt.show()